import re
import traceback
from abc import ABCMeta
from dataclasses import dataclass
from typing import List, Optional, Tuple, Union

from data_to_paper.exceptions import data_to_paperException

from .user_script_name import module_filename, get_gpt_module_frames


@dataclass
class AnyException(data_to_paperException):
    msg: str
    type_name: str

    def __str__(self):
        return self.msg


def convert_exception_to_any_exception_if_needed(e: Exception) -> Exception:
    if isinstance(e, (TimeoutError, SyntaxError, ImportError, RuntimeWarning, FileNotFoundError,
                      data_to_paperException)):
        return e
    from .overrides.pvalue import is_p_value, OnStrPValue, OnStr
    msg = None
    if isinstance(e, KeyError):
        key = e.args[0]
        if is_p_value(key):
            msg = 'Using PValue as a key in mapping is mot allowed.\nDo not attempt to map or transform PValue(s).'
    if msg is None:
        with OnStrPValue(OnStr.WITH_ZERO):
            msg = str(e)
    return AnyException(msg, type_name=type(e).__name__)


@dataclass
class FailedRunningCode(data_to_paperException):
    exception: Optional[Union[Exception, AnyException]] = None
    tb: Optional[traceback.StackSummary] = None
    py_spy_stack_and_code: Optional[Tuple[str, str]] = ('', '')
    fake_file_name = "my_analysis.py"

    @classmethod
    def from_exception(cls, e: Exception):
        return cls(exception=convert_exception_to_any_exception_if_needed(e), tb=traceback.extract_tb(e.__traceback__))

    @classmethod
    def from_current_tb(cls, **kwargs):
        exception = kwargs.pop('exception', None)
        return cls(exception=exception, tb=traceback.extract_stack(), **kwargs)

    @classmethod
    def from_exception_with_py_spy(cls, e: Exception, py_spy_stack_and_code: Tuple[str, str]):
        return cls(exception=convert_exception_to_any_exception_if_needed(e), tb=traceback.extract_tb(e.__traceback__),
                   py_spy_stack_and_code=py_spy_stack_and_code)

    def __str__(self):
        return f"Running the code resulted in the following exception:\n{self.exception}\n"

    @property
    def linenos_and_lines(self):
        return self.get_lineno_line_message()[0]

    def _get_gpt_module_frames(self):
        """
        returns the frame(s) of the module that was created by gpt.
        """
        if self.tb is None:
            return None
        return get_gpt_module_frames(self.tb)

    def _extract_linono_line_from_py_spy_stack(self):
        """
        returns the line of code that caused the exception.
        """
        search_results = re.search(fr'{module_filename}:(\d+)', self.py_spy_stack_and_code[0])
        if search_results is None:
            return []
        lineno = search_results.group(1)
        line = self.py_spy_stack_and_code[1].splitlines()[int(lineno) - 1]
        return [(lineno, line)]

    def get_lineno_line_message(self, lines_added_by_modifying_code: int = 0) -> Tuple[List[Tuple[int, str]], str]:
        """
        returns the line of code that caused the exception.
        """
        if isinstance(self.exception, SyntaxError):
            lineno = self.exception.lineno
            text = self.exception.text
            linenos_and_lines = [(lineno, text)]
            msg = self.exception.msg
        elif isinstance(self.exception, TimeoutError) and self.py_spy_stack_and_code[0]:
            msg = str(self.exception)
            linenos_and_lines = self._extract_linono_line_from_py_spy_stack()
            return linenos_and_lines, msg
        else:
            msg = str(self.exception)
            frames = self._get_gpt_module_frames()
            if frames is None:
                linenos_and_lines = []
            else:
                linenos_and_lines = [(f.lineno, f.line) for f in frames]
        linenos_and_lines = [(lineno - lines_added_by_modifying_code, text) for lineno, text in linenos_and_lines]
        return linenos_and_lines, msg

    def get_type_name(self):
        if isinstance(self.exception, AnyException):
            return self.exception.type_name
        return type(self.exception).__name__

    def get_traceback_message(self, lines_added_by_modifying_code: int = 0):
        """
        returns a fake traceback message, simulating as if the code ran in a real file.
        the line causing the exception is extracted from the ran `code`.
        """
        linenos_and_lines, msg = self.get_lineno_line_message(lines_added_by_modifying_code)
        s = ''
        for lineno, line in linenos_and_lines:
            s += f'  File "{self.fake_file_name}", line {lineno}, in <module>"\n' + \
               f'    {line}\n'
        return s + f'{self.get_type_name()}: {msg}'


class BaseRunContextException(data_to_paperException, metaclass=ABCMeta):
    pass


@dataclass
class CodeTimeoutException(BaseRunContextException, TimeoutError):
    time: int

    def __str__(self):
        return f"Code timeout after {self.time} seconds."


@dataclass
class UnAllowedFilesCreated(BaseRunContextException, PermissionError):
    un_allowed_files: List[str]

    def __str__(self):
        return f'UnAllowedFilesCreated: {self.un_allowed_files}'


@dataclass
class CodeUsesForbiddenFunctions(BaseRunContextException):
    func: str

    def __str__(self):
        return f"Code uses a forbidden function `{self.func}`."


@dataclass
class CodeImportForbiddenModule(BaseRunContextException):
    module: str

    def __str__(self):
        return f"Code import a forbidden module {self.module}."


@dataclass
class CodeWriteForbiddenFile(BaseRunContextException):
    file: str

    def __str__(self):
        return f"Code tries to create a forbidden file {self.file}."


@dataclass
class CodeReadForbiddenFile(BaseRunContextException):
    file: str

    def __str__(self):
        return f"Code tries to load a forbidden file {self.file}."



from dataclasses import dataclass

from data_to_paper.terminate.exceptions import TerminateException


@dataclass
class FailedCreatingProductException(TerminateException):
    reason: str

    def __str__(self):
        return self.reason



import time

import numpy as np
import requests
import re

from dataclasses import dataclass
from typing import List, Dict, Optional, Tuple

from data_to_paper.env import SEMANTIC_SCHOLAR_API_KEY

from data_to_paper.exceptions import data_to_paperException
from data_to_paper.latex.clean_latex import replace_special_latex_chars
from data_to_paper.utils.print_to_file import print_and_log_red
from data_to_paper.utils.nice_list import NiceList

from .base_server import ParameterizedQueryServerCaller
from .custom_types import Citation
from .types import (
    ServerErrorException,
    InvalidAPIKeyError,
    MissingSemanticScholarAPIKeyError,
)


# TODO: this is part of the WORKAROUND. remove it when the bug is fixed.
def remove_word(string, word):
    import re

    pattern = re.compile(
        pattern=r"\b{}\b\s*".format(re.escape(word)), flags=re.IGNORECASE
    )
    return re.sub(pattern, "", string)


PAPER_SEARCH_URL = "https://api.semanticscholar.org/graph/v1/paper/search"
EMBEDDING_URL = "https://model-apis.semanticscholar.org/specter/v1/invoke"


get_bibtex_id_from_bibtex = lambda bibtex: bibtex.split("{", 1)[1].split(",\n", 1)[0]


class SemanticCitation(Citation):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._bibtex_id = None

    @property
    def bibtex(self) -> str:
        bibtex = self["citationStyles"]["bibtex"]

        # remove non-ascii characters:
        bibtex = bibtex.encode("ascii", "ignore").decode("utf-8")
        bibtex = replace_special_latex_chars(bibtex)

        # remove commas from authors:
        try:
            authors = bibtex.split("author = {", 1)[1].split("},", 1)[0]
        except IndexError:
            pass
        else:
            bibtex = (
                bibtex.split("author = {", 1)[0]
                + "author = {"
                + authors
                + "},"
                + bibtex.split("},", 1)[1]
            )

        # characters not allowed in bibtex ids are replaced with '-':
        pattern = r"[{}(),\\\"-#~^:\'`ʹ_]"
        bibtex_id = get_bibtex_id_from_bibtex(bibtex)
        bibtex_id = re.sub(pattern, "-", bibtex_id)
        bibtex = (
            bibtex.split("{", 1)[0]
            + "{"
            + bibtex_id
            + ",\n"
            + bibtex.split(",\n", 1)[1]
        )

        return bibtex

    @property
    def bibtex_id(self) -> str:
        if self._bibtex_id is None:
            bibtex_id = get_bibtex_id_from_bibtex(self.bibtex)
            self._bibtex_id = bibtex_id
        return self._bibtex_id

    @property
    def title(self) -> Optional[str]:
        return self.get("title", None)

    @property
    def abstract(self) -> Optional[str]:
        return self.get("abstract", None)

    @property
    def journal(self) -> Optional[str]:
        try:
            return self["journal"]["name"]
        except (KeyError, TypeError):
            return None

    @property
    def year(self) -> Optional[str]:
        return self.get("year", None)

    @property
    def influence(self) -> int:
        return self["influentialCitationCount"]

    @property
    def embedding(self) -> Optional[np.ndarray]:
        return self.get("embedding", None)

    @property
    def tldr(self) -> Optional[str]:
        tldr = self.get("tldr", None)
        if tldr is None:
            return None
        return tldr["text"]


@dataclass
class ServerErrorNoMatchesFoundForQuery(data_to_paperException):
    """
    Error raised server wasn't able to find any matches for the query.
    """

    query: str

    def __str__(self):
        return f"Server wasn't able to find any matches for the query:\n {self.query}\n please try a different query."


class SemanticScholarPaperServerCaller(ParameterizedQueryServerCaller):
    """
    Search for citations with abstracts in Semantic Scholar.
    """

    name = "Semantic Scholar"
    file_extension = "_semanticscholar_paper.bin"

    @classmethod
    def _get_server_response(cls, query, rows=25) -> List[dict]:
        """
        Get the response from the semantic scholar server as a list of dict citation objects.
        """
        if SEMANTIC_SCHOLAR_API_KEY.key is None:
            missing_api_key_error = MissingSemanticScholarAPIKeyError(
                server=cls.name, api_key=SEMANTIC_SCHOLAR_API_KEY
            )
            print_and_log_red(str(missing_api_key_error), should_log=False)

        # TODO: THIS IS A WORKAROUND FOR A BUG IN SEMANTIC SCHOLAR. REMOVE WHEN FIXED.
        words_to_remove_in_case_of_zero_citation_error = (
            "the",
            "of",
            "in",
            "and",
            "or",
            "a",
            "an",
            "to",
            "for",
            "on",
            "at",
            "by",
            "with",
            "from",
            "as",
            "into",
            "through",
            "effect",
        )

        while True:
            params = {
                "query": query,
                "limit": min(
                    rows * 2, 100
                ),  # x2 more to make sure we get enough results after removing faulty ones
                "fields": "title,url,abstract,tldr,journal,year,citationStyles,embedding,influentialCitationCount",
            }
            print_and_log_red(
                f'QUERYING SEMANTIC SCHOLAR FOR: "{query}"', should_log=False
            )
            headers = (
                {"x-api-key": SEMANTIC_SCHOLAR_API_KEY.key}
                if SEMANTIC_SCHOLAR_API_KEY.key
                else {}
            )
            for attempt in range(3):
                response = requests.get(
                    PAPER_SEARCH_URL, headers=headers, params=params
                )
                if response.status_code not in (504, 429):
                    break
                wait_time = 2**attempt  # Exponential backoff
                print_and_log_red(
                    f"ERROR: Server timed out or too many requests. "
                    f"We wait for {wait_time} sec and try again.",
                    should_log=False,
                )
                time.sleep(wait_time)
            else:
                raise ServerErrorException(
                    server=cls.name, response=response
                )  # if we failed all attempts

            if response.status_code != 200:  # 200 is the success code
                if response.reason == "Forbidden":
                    raise InvalidAPIKeyError(
                        server=cls.name,
                        response=response,
                        api_key=SEMANTIC_SCHOLAR_API_KEY,
                    )
                raise ServerErrorException(server=cls.name, response=response)

            data = response.json()
            try:
                papers = data["data"]
            except KeyError:
                papers = []

            if len(papers) > 0:  # if there is no server bug
                papers = [
                    paper
                    for paper in papers
                    if SemanticCitation(paper).bibtex_id != "None"
                ]
                return papers[:rows]

            for word in words_to_remove_in_case_of_zero_citation_error:
                redacted_query = remove_word(query, word)
                if redacted_query != query:
                    print_and_log_red(
                        f"NO MATCHES!  REMOVING '{word}' FROM QUERY", should_log=False
                    )
                    query = redacted_query
                    break
            else:
                # failing gracefully
                return []

    @staticmethod
    def _get_embedding(paper: Dict[str, Dict[str, str]]) -> Tuple[np.ndarray, str]:
        """
        Get the embedding of the paper and a message if there was an error.
        """
        msg = ""
        embedding = None
        if "embedding" not in paper:
            msg = "No embedding attr"
        elif paper["embedding"] is None:
            msg = "None embedding attr"
        elif "model" not in paper["embedding"]:
            msg = "No model attr"
        elif paper["embedding"]["model"] not in ["specter@v0.1.1", "specter_v1"]:
            msg = "Wrong model attr"
        else:
            try:
                assert len(paper["embedding"]["vector"]) == 768
            except (AssertionError, KeyError, IndexError, TypeError):
                msg = "Wrong vector attr"
            else:
                embedding = np.array(paper["embedding"]["vector"])
        return embedding, msg

    @staticmethod
    def _post_process_response(response, args, kwargs):
        """
        Post process the response from the server.
        """
        query = args[0] if len(args) > 0 else kwargs.get("query", None)
        citations = NiceList(separator="\n", prefix="[\n", suffix="\n]")
        embedding_error_counts = {}
        for rank, paper in enumerate(response):
            embedding, msg = SemanticScholarPaperServerCaller._get_embedding(paper)
            paper = paper.copy()
            paper["embedding"] = embedding
            citation = SemanticCitation(paper, search_rank=rank, query=query)
            if len(citation.bibtex_id) <= 4:
                print_and_log_red(
                    f"ERROR: bibtex_id is too short. skipping. Title: {citation.title}"
                )
                continue
            if msg:
                embedding_error_counts[msg] = embedding_error_counts.get(msg, 0) + 1
            citations.append(citation)
        if embedding_error_counts:
            print_and_log_red(
                f"Total citations: {len(citations)}; {embedding_error_counts}"
            )
        return citations


class SemanticScholarEmbeddingServerCaller(ParameterizedQueryServerCaller):
    """
    Embed "paper" (title + abstract) using SPECTER Semantic Scholar API.
    """

    name = "Semantic Scholar Embedding"
    file_extension = "_semanticscholar_embedding.bin"

    @classmethod
    def _get_server_response(cls, paper: Dict[str, str]) -> np.ndarray:
        """
        Send the paper to the SPECTER Semantic Scholar API and get the embedding.
        """
        # check that the paper has id, title and abstract attributes, if not raise an error
        if not all(key in paper for key in ["paper_id", "title", "abstract"]):
            raise ValueError(
                "Paper must have 'paper_id', 'title' and 'abstract' attributes."
            )

        print_and_log_red(
            f'GETTING SEMANTIC SCHOLAR EMBEDDING FOR: "{paper["title"]}"',
            should_log=False,
        )
        response = requests.post(EMBEDDING_URL, json=[paper])

        if response.status_code != 200:
            raise ServerErrorException(server=cls.name, response=response)

        paper_embedding = response.json()["preds"][0]["embedding"]

        return np.array(paper_embedding)


SEMANTIC_SCHOLAR_SERVER_CALLER = SemanticScholarPaperServerCaller()
SEMANTIC_SCHOLAR_EMBEDDING_SERVER_CALLER = SemanticScholarEmbeddingServerCaller()



from __future__ import annotations

import re

from typing import List, Tuple, Optional, Set

from data_to_paper.utils.tag_pairs import SAVE_TAGS

from .message import Message, Role
from .message_designation import GeneralMessageDesignation, convert_general_message_designation_to_int_list

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from data_to_paper.base_cast import Agent


class Conversation(List[Message]):
    """
    Maintain a list of messages as exchanged between USER and ASSISTANT.

    Takes care of:

    1. save/load to text.
    2. print colored-styled messages.

    DO NOT ALTER CONVERSATION INSTANCE DIRECTLY. USE `ConversationManager` INSTEAD.
    """

    def __init__(self, *args, conversation_name: Optional[str] = None,
                 participants: Optional[Set[Agent]] = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.conversation_name = conversation_name
        self.participants = participants  # None - do not enforce participants

    def add_participant(self, agent: Agent):
        if self.participants is None:
            self.participants = []
        self.participants.add(agent)

    def remove_participant(self, agent: Agent):
        if self.participants is not None:
            self.participants.remove(agent)

    def get_other_participant(self, agent: Agent) -> Agent:
        if len(self.participants) != 2:
            raise ValueError('Conversation must have exactly two participants.')
        for participant in self.participants:
            if participant is not agent:
                return participant

    def append(self, message: Message):
        if self.participants is not None and message.role is not Role.COMMENTER:
            assert message.agent in self.participants, f'Agent {message.agent} not in conversation participants.'
        super().append(message)

    def get_chosen_indices_and_messages(self, hidden_messages: GeneralMessageDesignation = None
                                        ) -> List[Tuple[int, Message]]:
        """
        Return sub-list of messages.
        Remove commenter messages, ignore=True messages, as well as all messages indicated in `hidden_messages`.
        """
        hidden_messages = hidden_messages or []
        hidden_messages = convert_general_message_designation_to_int_list(hidden_messages, self)
        return [(i, message) for i, message in enumerate(self)
                if i not in hidden_messages
                and message.role is not Role.COMMENTER
                and not message.ignore]

    def get_chosen_messages(self, hidden_messages: GeneralMessageDesignation = None) -> List[Message]:
        """
        Return sub-list of messages.
        Remove commenter messages, ignore=True messages, as well as all messages indicated in `hidden_messages`.
        """
        return [message for _, message in self.get_chosen_indices_and_messages(hidden_messages)]

    def get_last_response(self) -> str:
        """
        Return the last response from the assistant.

        will skip over any COMMENTER messages.
        """
        last_non_commenter_message = self.get_last_non_commenter_message()
        if not last_non_commenter_message.role.is_assistant_or_surrogate():
            raise ValueError('Last message is not an assistant response.')
        return last_non_commenter_message.content

    def get_last_non_commenter_message(self) -> Message:
        """
        Return the last non-commenter message.
        """
        for i in range(len(self) - 1, -1, -1):
            if self[i].role is not Role.COMMENTER:
                return self[i]
        raise ValueError('No non-commenter message found.')

    def get_message_index_by_tag(self, tag):
        for i, message in enumerate(self):
            if message.tag == tag:
                return i
        raise ValueError(f'Tag {tag} not found.')

    def save(self, filename: str):
        with open(filename, 'w') as f:
            for message in self:
                f.write(SAVE_TAGS.wrap(message.convert_to_text()) + '\n\n')

    def load(self, filename: str):
        self.clear()
        with open(filename, 'r') as f:
            entire_file = f.read()
            matches = re.findall(SAVE_TAGS.wrap("(.*?)"), entire_file, re.DOTALL)
            for match in matches:
                self.append(Message.from_text(match))

    @classmethod
    def from_file(cls, filename: str):
        self = cls()
        self.load(filename)
        return self

    def print_all_messages(self):
        for message in self:
            print(message.pretty_repr())
            print()



import re

FROM_OPEN_BRACKET_TO_CLOSE_BRACKET = {'[': ']', '{': '}', '(': ')'}


def find_str(text: str, query: str, start: int = 0, end: int = None, case_sensitive: bool = True) -> int:
    if not case_sensitive:
        text = text.lower()
        query = query.lower()
    if end is None:
        end = len(text)
    return text.find(query, start, end)


def extract_text_between_most_flanking_tags(text: str, left_tag: str, right_tag: str = None, keep_tags: bool = False,
                                            case_sensitive: bool = True) -> str:
    """
    Extract text between the first left tag and the last right tag.
    """
    start = find_str(text, left_tag, case_sensitive=case_sensitive)
    if start == -1:
        raise ValueError(f'Could not find left tag {left_tag} in text')
    end = text.rfind(right_tag)
    if end == -1:
        raise ValueError(f'Could not find right tag {right_tag} in text')
    if keep_tags:
        return text[start:end + len(right_tag)]
    return text[start + len(left_tag):end]


def extract_text_between_tags(text: str, left_tag: str, right_tag: str = None, keep_tags: bool = False,
                              case_sensitive: bool = True) -> str:
    """
    Extract text between two tags.
    If the right tag is None, then extract text from the left tag to the end of the text
    We also take in account nested brackets.
    """
    optional_brackets = {'[': ']', '{': '}', '(': ')'}
    left_bracket = left_tag[-1]
    if right_tag is not None:
        right_bracket = right_tag[-1]
        if left_bracket not in optional_brackets.keys() or right_bracket != optional_brackets[left_bracket]:
            # just find the first instance of the right tag and return the text between the left tag and the right tag
            start = find_str(text, left_tag, case_sensitive=case_sensitive)
            if start == -1:
                raise ValueError(f'Could not find left tag {left_tag} in text')
            end = find_str(text, right_tag, start + len(left_tag), case_sensitive=case_sensitive)
            if end == -1:
                raise ValueError(f'Could not find left tag {right_tag} in text')
            if end - start - len(left_tag) == 0:
                raise ValueError(f'Could not find left tag {left_tag} in text')
            if keep_tags:
                return text[start:end + len(right_tag)]
            return text[start + len(left_tag):end]
        else:
            return extract_text_between_brackets(text, left_tag, keep_tags=keep_tags)
    else:
        # right tag is None, so we return the text from the left tag to the end of the text
        start = find_str(text, left_tag, case_sensitive=case_sensitive)
        if start == -1:
            raise ValueError(f'Could not find left tag {left_tag} in text')
        if keep_tags:
            return left_tag + text[start + len(left_tag):]
        return text[start + len(left_tag):]


def extract_text_between_brackets(text: str, open_tag: str, keep_tags: bool = False):
    """
    use stack to find matching closing bracket for the first open bracket, use stack to find matching closing bracket.
    return the text between the first open bracket and the matching closing bracket without the brackets.
    """
    start = text.find(open_tag)
    if start == -1:
        raise ValueError(f'Could not find open bracket {open_tag} in text')
    end = start + len(open_tag)
    open_bracket = open_tag[-1]
    close_bracket = FROM_OPEN_BRACKET_TO_CLOSE_BRACKET[open_bracket]
    stack = [open_bracket]
    while len(stack) > 0:
        if end == len(text):
            raise ValueError(f'Could not find matching closing bracket for open bracket {open_bracket} in text')
        if text[end] == open_bracket:
            stack.append(open_bracket)
        elif text[end] == close_bracket:
            stack.pop()
        end += 1
    if keep_tags:
        return text[start:end]
    return text[start + len(open_tag):end - 1]


def extract_all_external_brackets(text: str, open_bracket: str, close_bracket: str = None, open_phrase: str = None):
    """
    Extract all text between the open bracket and the matching closing bracket.
    For example, if open_bracket is '[', and text is 'hello [world [inner]], what is your [name]', then return
    ['[world [inner]]', '[name]'].
    if there are no open brackets, return an empty list.
    """
    if close_bracket is None:
        close_bracket = FROM_OPEN_BRACKET_TO_CLOSE_BRACKET[open_bracket]
    if open_phrase is None:
        open_phrase = open_bracket
    start = text.find(open_phrase)
    if start == -1:
        return []
    end = start + len(open_phrase)
    stack = [open_bracket]
    while len(stack) > 0:
        if end == len(text):
            return []
        if text[end] == open_bracket:
            stack.append(open_bracket)
        elif text[end] == close_bracket:
            stack.pop()
        end += 1
    return [text[start:end]] + extract_all_external_brackets(text[end:], open_bracket, close_bracket, open_phrase)


def _extract_to_nearest(text: str, max_length: int, char: str = '\n'):
    """
    Extract the text from the beginning of the text to the nearest char before end.
    If no char is found, extract the text from the beginning of the text to end.
    if max_length is negative, extract the text from the end of the text to the nearest char before end.
    """
    if abs(max_length) > len(text):
        return text
    if max_length >= 0:
        text = text[:max_length]
        end = text.rfind(char) if char in text else max_length
        return text[:end]
    else:
        text = text[max_length:]
        end = text.find(char) + 1 if char in text else max_length
        return text[end:]


def extract_to_nearest_newline(text: str, end: int):
    """
    Extract the text from the beginning of the text to the nearest newline before end.
    If no newline is found, extract the text from the beginning of the text to end.
    """
    return _extract_to_nearest(text, end, '\n')


def extract_to_nearest_space(text: str, end: int):
    """
    Extract the text from the beginning of the text to the nearest space before end.
    If no space is found, extract the text from the beginning of the text to end.
    """
    return _extract_to_nearest(text, end, ' ')


def get_dot_dot_dot_text(text: str, start: int, end: int):
    """
    Get the text from the beginning of the text to the nearest space before start and from the nearest space after end
    to the end of the text.
    """
    fill = ' [...] '
    start -= len(fill)
    text = re.sub(' +', ' ', text)
    text = text.replace('\n', ' ').replace('```', '').strip()
    if start - end + len(fill) > len(text):
        return text
    return extract_to_nearest_space(text, start) + fill + extract_to_nearest_space(text, end)



from dataclasses import dataclass, field
from pathlib import Path
from typing import Tuple, Optional

from data_to_paper.base_steps import BaseCodeProductsGPT
from data_to_paper.conversation.actions_and_conversations import ActionsAndConversations
from data_to_paper.llm_coding_utils import ALLOWED_PLOT_KINDS
from data_to_paper.research_types.hypothesis_testing.cast import ScientificAgent
from data_to_paper.research_types.hypothesis_testing.scientific_products import ScientificProducts, get_code_name, \
    get_code_agent
from data_to_paper.text import dedent_triple_quote_str
from data_to_paper.utils.nice_list import NiceList


@dataclass
class BaseScientificCodeProductsHandler:
    code_step: str = ''  # "data_analysis", "data_exploration", "data_processing"
    products: ScientificProducts = None
    assistant_agent: ScientificAgent = ScientificAgent.Performer
    user_agent: ScientificAgent = None

    actions_and_conversations: ActionsAndConversations = field(default_factory=ActionsAndConversations)
    code_name: str = None
    conversation_name: str = None

    def __post_init__(self):
        if self.code_name is None:
            self.code_name = get_code_name(self.code_step)
        if self.conversation_name is None:
            self.conversation_name = f'{self.code_name} Code'
        if self.user_agent is None:
            self.user_agent = get_code_agent(self.code_step)


@dataclass
class BaseScientificCodeProductsGPT(BaseScientificCodeProductsHandler, BaseCodeProductsGPT):
    allow_data_files_from_sections: Tuple[Optional[str]] = (None, )  # None for the raw data files, () for no data files
    background_product_fields: Tuple[str, ...] = ('all_file_descriptions', 'research_goal')

    def __post_init__(self):
        BaseScientificCodeProductsHandler.__post_init__(self)
        BaseCodeProductsGPT.__post_init__(self)

    @property
    def files_available_from_prior_stages(self) -> NiceList[str]:
        files = NiceList([], wrap_with='"', separator='\n')
        for section in self.allow_data_files_from_sections:
            if section is None:
                continue
            if section in self.products.codes_and_outputs:
                files += \
                    self.products.codes_and_outputs[section].created_files.get_created_files_available_for_next_steps()
        return files

    @property
    def data_filenames(self) -> NiceList[str]:
        return NiceList(self.raw_data_filenames + self.files_available_from_prior_stages,
                        wrap_with='"', prefix='\n', separator='\n', suffix='\n')

    @property
    def list_additional_data_files_if_any(self) -> str:
        if len(self.files_available_from_prior_stages) == 0:
            return ''
        return f'\nOr you can also use the processed files created above by the data processing code:\n' \
               f'```\n' \
               f'{self.files_available_from_prior_stages}' \
               f'```\n' \
               f'Important: use the correct version of the data to perform each of the steps. For example, ' \
               f'for descriptive statistics use the original data, for model building use the processed data.'

    @property
    def raw_data_filenames(self) -> NiceList[str]:
        if None in self.allow_data_files_from_sections:
            return NiceList(self.products.data_file_descriptions.get_data_filenames(),
                            wrap_with='"',
                            prefix='{} data file[s]: ')
        return NiceList([], wrap_with='"', prefix='No data files.')

    @property
    def data_folder(self) -> Optional[Path]:
        return Path(self.products.data_file_descriptions.data_folder)


@dataclass
class BaseTableCodeProductsGPT(BaseScientificCodeProductsGPT):
    df_to_latex_extra_vars: str = dedent_triple_quote_str('''
        note: str = None, 
        glossary: Dict[Any, str] = None,
        ''', indent=8)

    df_to_latex_extra_vars_explain: str = dedent_triple_quote_str('''
        `note`: Note to be added below the table caption.
        `glossary`: Glossary for the table.
        ''', indent=4)

    df_to_latex_doc: str = dedent_triple_quote_str('''
        def df_to_latex(df, 
                filename: str, caption: str,
        {df_to_latex_extra_vars}\t
            ):
            """
            Saves a DataFrame `df` and creates a LaTeX table.
            `filename`, `caption`: as in `df.to_latex`.
        {df_to_latex_extra_vars_explain}\t
            """
        ''')

    df_to_figure_extra_latex_vars: str = dedent_triple_quote_str('''
        note: str = None, glossary: Dict[Any, str] = None,
        ''', indent=8)

    df_to_figure_extra_plot_vars: str = dedent_triple_quote_str('''
        xlabel: str = None, ylabel: str = None,
        ''', indent=8)

    df_to_figure_extra_latex_vars_explain: str = dedent_triple_quote_str('''
        `note`: Note to be added below the figure caption.
        `glossary`: Glossary for the figure.
        ''', indent=4)

    df_to_figure_extra_plot_vars_explain: str = dedent_triple_quote_str('''
        `xlabel`: Label for the x-axis.
        `ylabel`: Label for the y-axis.
        ''', indent=4)

    df_to_figure_doc: str = dedent_triple_quote_str('''
        def df_to_figure(
                df, filename: str, caption: str,
        {df_to_figure_extra_latex_vars}\t
                x: Optional[str] = None, y: List[str] = None, 
                kind: str = 'bar',
                logx: bool = False, logy: bool = False,
                y_ci: Optional[List[str]] = None,
                y_p_value: Optional[List[str]] = None,
        {df_to_figure_extra_plot_vars}\t
            ):
            """
            Save a DataFrame `df` and create a LaTeX figure.
            Parameters, for LaTex embedding of the figure:
            `filename`: Filename for the figure.
            `caption`: Caption for the figure.
        {df_to_figure_extra_latex_vars_explain}\t

            Parameters for df.plot():
            `x`: Column name for x-axis (index by default).
            `y`: List of m column names for y-axis (m=1 for single plot, m>1 for multiple plots).
            `kind`: {allowed_plot_kinds}.
            `logx` / `logy` (bool): log scale for x/y axis.
        {df_to_figure_extra_plot_vars_explain}\t

            `y_ci`: Confidence intervals for errorbars. 
                List of m column names indicating confidence intervals for each y column. 
                Each element in these columns must be a Tuple[float, float], \t
        describing the lower and upper bounds of the CI. 

             `y_p_value`: List of m column names (List[str]) containing numeric p-values of the \t
        corresponding y columns. These numeric values will be automatically converted by df_to_figure to \t
        stars ('***', '**', '*', 'ns') and plotted above the error bars.

            If provided, the length of `y_ci`, and `y_p_value` should be the same as of `y`.

            Example:
            Suppose, we have:

            df_lin_reg_longevity = pd.DataFrame({
                'adjusted_coef': [0.4, ...], 'adjusted_coef_ci': [(0.35, 0.47), ...], \t
        'adjusted_coef_pval': [0.012, ...],   
                'unadjusted_coef': [0.2, ...], 'unadjusted_coef_ci': [(0.16, 0.23), ...], \t
        'unadjusted_coef_pval': [0.0001, ...],
            }, index=['var1', ...])

            then:
            df_to_figure(df_lin_reg_longevity, 'df_lin_reg_longevity', caption='Coefficients of ...', kind='bar',  
                y=['adjusted_coef', 'unadjusted_coef'], 
                y_ci=['adjusted_coef_ci', 'unadjusted_coef_ci'], 
                y_p_value=['adjusted_coef_pval', 'unadjusted_coef_pval'])
            """
        ''')

    @property
    def allowed_plot_kinds(self) -> str:
        if len(ALLOWED_PLOT_KINDS) == 1:
            return f'only {ALLOWED_PLOT_KINDS[0]} is allowed'
        return str(NiceList(ALLOWED_PLOT_KINDS, wrap_with="'", separator=', ', last_separator=' or '))



from __future__ import annotations

import difflib
import colorama

from dataclasses import dataclass
from enum import Enum
from typing import NamedTuple, Optional, List

from data_to_paper.env import TEXT_WIDTH, MINIMAL_COMPACTION_TO_SHOW_CODE_DIFF, HIDE_INCOMPLETE_CODE, SHOW_LLM_CONTEXT
from data_to_paper.base_cast import Agent
from data_to_paper.run_gpt_code.code_utils import extract_code_from_text, FailedExtractingBlock
from data_to_paper.servers.llm_call import count_number_of_tokens_in_message
from data_to_paper.servers.model_engine import OpenaiCallParameters, ModelEngine
from data_to_paper.text import line_count, wrap_as_block
from data_to_paper.text.highlighted_text import colored_text, format_text_with_code_blocks
from data_to_paper.utils.numerics import is_lower_eq
from data_to_paper.text.formatted_sections import FormattedSections
from data_to_paper.text.text_extractors import get_dot_dot_dot_text

# noinspection PyUnresolvedReferences
colorama.just_fix_windows_console()


class Role(Enum):
    SYSTEM = 'system'
    USER = 'user'
    ASSISTANT = 'assistant'
    SURROGATE = 'surrogate'
    COMMENTER = 'commenter'

    def is_assistant_or_surrogate(self):
        return self in [Role.ASSISTANT, Role.SURROGATE]

    def is_not_commenter(self):
        return self is not Role.COMMENTER


class ResponseStyle(NamedTuple):
    color: str
    separator: str


ROLE_TO_STYLE = {
    Role.SYSTEM: ResponseStyle(colorama.Fore.GREEN, '-'),
    Role.USER: ResponseStyle(colorama.Fore.GREEN, '-'),
    Role.ASSISTANT: ResponseStyle(colorama.Fore.CYAN, '='),
    Role.SURROGATE: ResponseStyle(colorama.Fore.CYAN, '='),
    Role.COMMENTER: ResponseStyle(colorama.Fore.BLUE, ' '),
}


@dataclass
class Message:
    role: Role
    content: str
    tag: str = ''
    agent: Optional[Agent] = None
    openai_call_parameters: Optional[OpenaiCallParameters] = None
    ignore: bool = False  # if True, this message will be skipped when calling openai
    is_background: bool = False
    # True: message will not be shown in the web conversation
    # False: message will be shown in web conversation
    # None: message will be shown only if not repeated

    index_in_conversation: Optional[int] = None
    # index of the message in the conversation

    effective_index_in_conversation: Optional[int] = None
    # index of the message in the conversation, ignoring commenter and ignored messages

    context: List[Message] = None

    def to_llm_dict(self):
        return {'role': Role.ASSISTANT.value if self.role.is_assistant_or_surrogate()
                else self.role.value, 'content': self.content}

    def get_llm_model(self):
        if self.openai_call_parameters is None:
            return None
        return self.openai_call_parameters.model_engine

    def pretty_repr(self, number: Optional[int] = None, conversation_name: str = '', is_color: bool = True,
                    abbreviate_content: bool = False) -> str:
        """
        Returns a pretty repr of the message with color and heading.

        number: message sequential number in the conversation.

        Takes care of:
        * Indicating Role with name and color.
        * Adding separation lines
        * Indenting text
        * Highlighting code blocks
        """
        role, content, tag, agent = self.role, self.content, self.tag, self.agent
        tag_text = f' <{tag}> ' if tag else ''
        agent_text = f' {{{agent.value}}}' if agent else ''
        num_text = f'[{number}] ' if number else ''
        style = ROLE_TO_STYLE[role]
        sep = style.separator
        text_color = style.color if is_color else ''

        role_text = role.name + (f'{self.openai_call_parameters}' if self.openai_call_parameters else '')
        if role == Role.SYSTEM:
            role_model_agent_conversation_tag = f'{role_text} casting {agent_text} -> "{conversation_name}" '
        else:
            role_model_agent_conversation_tag = f'{role_text}{agent_text} -> "{conversation_name}" {tag_text} '

        if role == Role.COMMENTER:
            return colored_text(num_text + role_model_agent_conversation_tag + ': ' + content, text_color)

        if abbreviate_content:
            start = TEXT_WIDTH * 7 // 10
            content = get_dot_dot_dot_text(content, start=start, end=start - TEXT_WIDTH)
            return colored_text(num_text + role_model_agent_conversation_tag + ': \n' + content, text_color)

        # header:
        s = colored_text(num_text + sep * (9 - len(num_text)) + ' ' + role_model_agent_conversation_tag
                         + sep * (TEXT_WIDTH - len(role_model_agent_conversation_tag) - 9 - 1) + '\n', text_color)

        # content:
        s += self.pretty_content(text_color, width=TEXT_WIDTH)
        if s[-1] != '\n':
            s += '\n'

        # footer:
        s += colored_text(sep * TEXT_WIDTH, text_color)
        return s

    def get_content_after_hiding_incomplete_code(self) -> (str, bool):
        """
        Detect if the message contains incomplete code.
        """
        content = self.content.strip()
        if not content:
            return content, False
        formatted_sections = FormattedSections.from_text(content)
        if len(formatted_sections) == 0:
            is_incomplete_code = False
        else:
            last_section = formatted_sections.get_last_block()
            is_incomplete_code = HIDE_INCOMPLETE_CODE and last_section is not None and not last_section.is_complete
        return content, is_incomplete_code

    def get_number_of_tokens(self, model_engine: ModelEngine = None) -> int:
        model_engine = model_engine or self.get_llm_model()
        return count_number_of_tokens_in_message([self], model_engine)

    def get_number_of_tokens_in_context(self) -> int:
        if self.context is None:
            return 0
        return count_number_of_tokens_in_message(self.context, self.get_llm_model())

    def _get_triple_quote_formatted_content(self, with_header: bool = True) -> (str, bool):
        content, is_incomplete_code = self.get_content_after_hiding_incomplete_code()
        if self.role == Role.SYSTEM:
            content = wrap_as_block(content, 'system')
        if self.role == Role.COMMENTER:
            content = wrap_as_block(content, 'comment')
        if self.effective_index_in_conversation is not None:
            index = self.effective_index_in_conversation
        else:
            index = len(self.context) if self.context else None

        if SHOW_LLM_CONTEXT and with_header and self.role != Role.COMMENTER and index is not None:
            llm_parameters = f'{self.openai_call_parameters}' if self.openai_call_parameters else ''
            header = ''
            if self.context:
                header += f'CONTEXT TOTAL ({self.get_number_of_tokens_in_context()} tokens):\n'
                for i, message in enumerate(self.context):
                    header += f'#{i:>2} {message.get_short_description(model=self.get_llm_model())}\n'
            header += f'\n#{index:>2} {self.get_short_description()}\n' + ' ' * 79 + \
                      f'{llm_parameters}'
            header = wrap_as_block(header, 'header')
            content = header + '\n\n' + content

        return content, is_incomplete_code

    def _make_content_for_pretty(self, content: str, is_incomplete_block: bool):
        return content

    def pretty_content(self, text_color: str = '', width: Optional[int] = None, is_html=False, with_header: bool = True
                       ) -> str:
        """
        Returns a pretty repr of just the message content.
        """
        content, is_incomplete_code = self._get_triple_quote_formatted_content(with_header)
        content = self._make_content_for_pretty(content, is_incomplete_code)
        return format_text_with_code_blocks(text=content, text_color=text_color, width=width, is_html=is_html,
                                            from_md=None, do_not_format=['latex'])

    def get_short_description(self, left: int = 85, right: int = -20, model: ModelEngine = None) -> str:
        return f'{self.role.name:>9} ({self.get_number_of_tokens(model):>4} tokens): ' \
               f'{get_dot_dot_dot_text(self.content, left, right)}'

    def convert_to_text(self):
        return f'{self.role.value}<{self.tag}>\n{self.content}'

    @classmethod
    def from_text(cls, text):
        first_break = text.index('\n')
        first_lt = text.index('<')
        role = Role(text[:first_lt])
        tag = text[first_lt + 1: first_break - 1]
        content = text[first_break + 1:]
        return cls(role=role, content=content, tag=tag)


@dataclass
class CodeMessage(Message):
    """
    A message that contains code that needs to be compared to some previous code.
    """

    previous_code: str = None
    "The code from the previous response, to which this code should be compared"

    @property
    def extracted_code(self) -> Optional[str]:
        """
        Extract the code from the response.
        """
        try:
            return extract_code_from_text(self.content)
        except FailedExtractingBlock:
            return None

    def get_code_diff(self) -> str:
        """
        Get the difference between the code from the previous response and the code from this response.
        """
        diff = difflib.unified_diff(self.previous_code.strip().splitlines(),
                                    self.extracted_code.strip().splitlines(),
                                    lineterm='', n=0)
        # we remove the first 3 lines, which are the header of the diff:
        diff = list(diff)[3:]
        return '\n'.join(diff)

    def get_content_after_hiding_incomplete_code(self) -> (str, bool):
        """
        Detect if the message contains incomplete code.
        """
        content, is_incomplete_code = super().get_content_after_hiding_incomplete_code()
        if is_incomplete_code:
            formatted_sections = FormattedSections.from_text(content)
            last_section = formatted_sections.get_last_block()
            partial_code = last_section.section
            last_section.section = \
                f"\n# NOT SHOWING INCOMPLETE CODE SENT BY LLM ({line_count(partial_code)} LINES)\n"
            last_section.is_complete = True
            content = formatted_sections.to_text()
        return content, is_incomplete_code

    def _make_content_for_pretty(self, content: str, is_incomplete_block: bool):
        """
        We override this method to replace the code within the message with the diff.
        """
        if self.extracted_code and not is_incomplete_block and self.previous_code:
            diff = self.get_code_diff()
            if not is_lower_eq(line_count(self.extracted_code) - line_count(diff),
                               MINIMAL_COMPACTION_TO_SHOW_CODE_DIFF):
                # if the code diff is substantially shorter than the code, we replace the code with the diff:
                content = content.replace(
                    self.extracted_code,
                    "\n# FULL CODE SENT BY LLM IS SHOWN AS A DIFF WITH PREVIOUS CODE\n" + diff if diff
                    else "\n# LLM SENT THE SAME CODE AS BEFORE\n")
        return content


@dataclass
class JsonMessage(Message):
    def get_content_after_hiding_incomplete_code(self) -> (str, bool):
        return wrap_as_block(self.content, 'json'), False


def create_message(role: Role, content: str, tag: str = '', agent: Optional[Agent] = None, ignore: bool = False,
                   openai_call_parameters: OpenaiCallParameters = None, context: List[Message] = None,
                   previous_code: str = None, is_code: bool = False,
                   is_json: bool = False,
                   is_background: bool = False) -> Message:
    kwargs = dict(role=role, content=content, tag=tag, agent=agent, ignore=ignore,
                  openai_call_parameters=openai_call_parameters, context=context,
                  is_background=is_background)

    is_code = is_code or previous_code is not None
    if is_json:
        return JsonMessage(**kwargs)
    if is_code:
        return CodeMessage(previous_code=previous_code, **kwargs)
    return Message(**kwargs)


def create_message_from_other_message(other_message: Message,
                                      content: Optional[str] = None,
                                      agent: Optional[Agent] = None) -> Message:
    return create_message(role=other_message.role,
                          content=content if content else other_message.content,
                          tag=other_message.tag,
                          agent=agent if agent else other_message.agent,
                          ignore=other_message.ignore,
                          openai_call_parameters=other_message.openai_call_parameters,
                          context=other_message.context,
                          is_code=isinstance(other_message, CodeMessage),
                          previous_code=other_message.previous_code if isinstance(other_message, CodeMessage) else None,
                          is_background=other_message.is_background)



# Utils made available to the LLM code for the display items step

from .df_to_latex import displayitems_df_to_latex as df_to_latex
from .df_to_figure import displayitems_df_to_figure as df_to_figure
from .df_formatting_utils import is_str_in_df, split_mapping, AbbrToNameDef
df_to_latex.__qualname__ = 'df_to_latex'
df_to_figure.__qualname__ = 'df_to_figure'



from dataclasses import dataclass
from typing import NamedTuple, Union, Callable, Tuple, Dict, List, Any

from data_to_paper.base_products.product import Product
from data_to_paper.conversation.stage import Stage
from data_to_paper.text.highlighted_text import format_text_with_code_blocks
from data_to_paper.text.text_formatting import ArgsOrKwargs, format_with_args_or_kwargs


class NameDescriptionStage(NamedTuple):
    name: str
    description: str
    stage: Stage


class NameDescriptionStageGenerator(NamedTuple):
    name: str
    description: str
    stage: Union[Stage, Callable]
    func: Callable


class ProductGenerator(NamedTuple):
    product: Union[Product, Callable]
    kwargs: Union[Dict[str, Any], Callable]


UnifiedProduct = Union[NameDescriptionStage, Product]
UnifiedProductGenerator = Union[NameDescriptionStageGenerator, ProductGenerator]


def _convert_args_or_kwargs_to_args(args_or_kwargs: ArgsOrKwargs) -> Tuple[str]:
    """
    Convert the given args or kwargs to args.
    """
    if isinstance(args_or_kwargs, tuple):
        return args_or_kwargs
    else:
        return tuple(args_or_kwargs.values())


@dataclass
class Products:
    """
    Contains the different outcomes of the process.
    These outcomes are gradually populated, where in each stage we get a new product based on previous products
    from prior stages.
    """

    _fields_to_unified_product_generators: Dict[str, UnifiedProductGenerator] = None
    _raise_on_none: bool = False

    def __post_init__(self):
        self._fields_to_unified_product_generators = self._get_generators()

    def _get_generators(self) -> Dict[str, UnifiedProductGenerator]:
        """
        Return a dictionary mapping product fields to a tuple of
        (name: str, description: str, stage: Stage, func: Callable).
        func is a function that creates args for the name and description to be formatted with.
        """
        return {}

    def get_name(self, product_field: str) -> str:
        """
        Return the name of the given product.
        """
        return self._get_name_description_stage(product_field).name

    def get_description(self, product_field: str) -> str:
        """
        Return the description of the given product.
        """
        return self._get_name_description_stage(product_field).description

    def get_stage(self, product_field: str) -> Stage:
        """
        Return the stage of the given product.
        """
        return self._get_name_description_stage(product_field).stage

    @staticmethod
    def extract_subfields(field: str) -> List[str]:
        """
        Return a list of subfields of the given field.
        """
        return field.split(':')

    def _get_unified_product_and_variables(self, field: str
                                           ) -> Tuple[UnifiedProduct, ArgsOrKwargs]:
        """
        Return the name, stage, and description variables of the given field.
        """
        unified_product_generator, args = self._get_unified_product_generator_and_args(field)
        if isinstance(unified_product_generator, NameDescriptionStageGenerator):
            name, description, stage, func = unified_product_generator
            variables = func(*args)
            if not isinstance(stage, Stage):
                stage = stage(*args)
            if not isinstance(variables, (tuple, dict)):
                variables = (variables, )
            return NameDescriptionStage(name, description, stage), variables
        elif isinstance(unified_product_generator, ProductGenerator):
            product, kwargs = unified_product_generator
            if not isinstance(product, Product):
                product = product(*args)
            if not isinstance(product, Product):
                raise ValueError(f'Unknown product field: {field}')
            if not product.is_valid():
                raise ValueError(f'Product {product} is not valid')
            if not isinstance(kwargs, dict):
                kwargs = kwargs(*args)
            return product, kwargs
        else:
            raise ValueError(f'Unknown product field: {field}')

    def _get_name_description_stage(self, field: str) -> NameDescriptionStage:
        """
        Return the name, stage, and description generator of the given field.
        """
        unified_product, variables = self._get_unified_product_and_variables(field)
        if self._raise_on_none and any(v is None for v in _convert_args_or_kwargs_to_args(variables)):
            raise ValueError(f'One of the variables in {variables} is None')
        if isinstance(unified_product, NameDescriptionStage):
            name, description, stage = unified_product
            name = format_with_args_or_kwargs(name, variables)
            description = format_with_args_or_kwargs(description, variables)
            return NameDescriptionStage(name, description, stage)
        elif isinstance(unified_product, Product):
            format_name = variables.pop('format_name') \
                if 'format_name' in variables else 'markdown'
            level = variables.pop('level') if 'level' in variables else 2
            return NameDescriptionStage(
                unified_product.get_header(**variables),
                unified_product.as_formatted_text(with_header=False, **variables),
                unified_product.get_stage(format_name=format_name, level=level,
                                          **variables),
            )
        else:
            raise ValueError(f'Unknown product field: {field}')

    def get_description_as_html(self, field: str) -> str:
        """
        Return the product of the given field.
        """
        unified_product, variables = self._get_unified_product_and_variables(field)
        if isinstance(unified_product, Product):
            return unified_product.as_html(level=1, **variables)
        description = self.get_description(field)
        return ('<h1>' + self.get_name(field) + '</h1>\n' +
                format_text_with_code_blocks(description, is_html=True, width=None, from_md=True))

    def get_description_for_llm(self, field: str) -> str:
        """
        Return the product in a format to be used in conversation context for the LLM.
        """
        name = self.get_name(field)
        description = self.get_description(field)
        return f'# {name}\n{description}'

    def _get_unified_product_generator_and_args(self, field: str
                                                ) -> Tuple[UnifiedProductGenerator, List[str]]:
        """
        Return the name, stage, and description of the given field.
        """
        subfields = self.extract_subfields(field)
        for current_field, unified_product in self._fields_to_unified_product_generators.items():
            current_subfields = self.extract_subfields(current_field)
            wildcard_subfields = []
            if len(subfields) == len(current_subfields):
                for subfield, current_subfield in zip(subfields, current_subfields):
                    if current_subfield == '{}':
                        wildcard_subfields.append(subfield)
                    elif subfield != current_subfield:
                        break
                else:
                    return unified_product, wildcard_subfields
        raise ValueError(f'Unknown product field: {field}')

    def is_product_available(self, field: str) -> bool:
        """
        Return whether the given product field is available.
        A product field is available if all of its variables are not None and all of the sub-products it
        depends on are available (not None).
        Trying to access an unavailable attributes is also interpreted as False, namely the sub-product
        is not available.
        """
        try:
            self._raise_on_none = True
            _, variables = self._get_unified_product_and_variables(field)
            variables = _convert_args_or_kwargs_to_args(variables)
            return all(variable is not None for variable in variables)
        except (KeyError, AttributeError, ValueError):
            return False
        finally:
            self._raise_on_none = False

    def __getitem__(self, item) -> NameDescriptionStage:
        return self._get_name_description_stage(item)



import re
from typing import List, Optional, Tuple


def extract_numeric_values(text: str) -> List[str]:
    """
    Extract all the numeric values from the given text.
    """
    # use regex to extract all the numeric values:
    return re.findall(r"[+-]?\d+(?:\.\d+)?(?:[eE][+-]?\d+)?", text.replace('{,}', '').replace(',', ''))


def unify_representation_of_numeric_values(text: str) -> str:
    """
    Unify the representation of numeric values with scientific notation.
    For example:
    "4.32 \times 10^{-5}" -> 4.32e-5
    "4.32 \times 10^5" -> 4.32e5
    "23.7987 * 10^5" -> 23.7987e5
    "23.7987*10^{-5}" -> 23.7987e-5
    """

    # This function is used to format the matched groups into scientific notation
    def repl(m):
        base = float(m.group(1))
        exponent = int(m.group(2).replace('{', '').replace('}', ''))  # remove curly braces
        return f'{base}e{exponent}'

    # Create regex pattern for numbers in the specified format
    pattern = r"(\d+\.\d+)\s*\\times\s*10\^(\{?\-?\d+\}?)"  # \times version
    text = re.sub(pattern, repl, text)

    pattern = r"(\d+\.\d+)\s*\*\s*10\^(\{?\-?\d+\}?)"  # * version
    text = re.sub(pattern, repl, text)

    return text


def is_one_with_zeros(str_number: str) -> bool:
    """
    Check if the given string number is 1 with zeros before or after.
    Like 0.001, 0.1, 1, 10, 100, etc.
    """
    return str_number.lstrip('-').replace(',', '').replace('.', '').lstrip('0').rstrip('0') == '1'


def is_int_below_max(str_number: str, max_int: int) -> bool:
    """
    Check if the given string number is an int below the given max int.
    """
    return '.' not in str_number and ',' not in str_number and 'e' not in str_number.lower() \
        and abs(int(str_number)) < max_int


def round_to_n_digits(str_number: str, n_digits: int, remove_sign: bool = True) -> float:
    """
    Round the given number to the given number of digits.
    """
    number = float(str_number.replace(',', ''))
    rounded = float(f'{float(f"{number:.{n_digits}g}"):g}')
    if remove_sign:
        return abs(rounded)
    return rounded


def truncate_to_n_digits(str_number: str, n_digits: int, remove_sign: bool = True) -> float:
    """
    Truncate the given number to the given number of digits.
    1.237 -> 1.23  (n_digits=3)
    """
    str_number = str_number.replace(',', '')
    str_number, power = split_number_and_power(str_number)
    digit_count = 0
    is_leading_zero = True
    is_after_point = False
    for i in range(len(str_number)):
        digit = str_number[i]
        if digit == '-' or digit == '+':
            continue
        if digit == '.':
            is_after_point = True
            continue
        if is_leading_zero and digit != '0':
            is_leading_zero = False
        if not is_leading_zero:
            digit_count += 1
        if digit_count == n_digits:
            break
    if not is_after_point:
        power = power + len(str_number) - i - 1
    truncated = float(str_number[:i + 1]) * 10 ** power
    if remove_sign:
        return abs(truncated)
    return truncated


def is_after_smaller_than_sign(str_number: str, target: str) -> Optional[bool]:
    """
    Check if the given string number extracted from the target str appear after a '<' sign.
    """
    str_number_positions = [m.start() for m in re.finditer(re.escape(str_number), target)]
    for str_number_position in str_number_positions:
        if str_number_position > 0 and target[str_number_position - 1] == '<' or \
                str_number_position > 1 and target[str_number_position - 2] == '<':
            return True
    return False


def is_percentage(str_number: str, target: str, search_distance: int = 30) -> Optional[bool]:
    """
    Check if the given string number extracted from the target str is a percentage.
    True: if the string matches a string in the target the ends with '%' (in the target).
    Nane: (maybe percentage) if the word 'percent'/'percentage' appears up to search_distance characters before/after
    the string number.
    False: otherwise
    """
    target_words = target.split()
    # find all the occurrences of the string number in the target:
    str_number_positions = [m.start() for m in re.finditer(re.escape(str_number), target)]
    len_str_number = len(str_number)

    # check if the string number is a percentage:
    for str_number_position in str_number_positions:
        # check if the string number is a percentage (ends with '%'):
        if str_number_position + len_str_number < len(target) and target[str_number_position + len_str_number] == '%':
            return True
        # check if the string number is a maybe percentage
        # (the word 'percent'/'percentage' appears up to search_distance characters before/after the string number):
        for keyword in ['percent', 'percentage']:
            if keyword in target_words[max(0, str_number_position - search_distance):
                                       str_number_position + len_str_number + search_distance]:
                return None
    return False


def is_any_matching_value_after_rounding_to_n_digits(source_str_numbers: List[str], target_number: float,
                                                     n_digits: int) -> bool:
    """
    Check if there exists a number in the source that matches after rounding to the given number of digits.
    """
    any_match = \
        any(round_to_n_digits(source_number, n_digits) == target_number
            for source_number in source_str_numbers)
    # if a number ends with '5' we allow also rounding it upwards
    any_match = any_match or \
        any(round_to_n_digits(source_number[:-1] + '6', n_digits) == target_number
            for source_number in source_str_numbers if source_number.endswith('5'))
    return any_match


def is_any_matching_value_after_truncating_to_n_digits(source_str_numbers: List[str], target_number: float,
                                                       n_digits: int) -> bool:
    """
    Check if there exists a number in the source that matches after rounding to the given number of digits.
    """
    return any(truncate_to_n_digits(source_number, n_digits) == target_number
               for source_number in source_str_numbers)


def get_number_of_significant_figures(str_number: str, remove_trailing_zeros: bool = True) -> int:
    """
    Get the number of significant figures in the given string number.
    """
    digits = str_number.replace('.', '').replace(',', '').replace('-', '')
    # remove leading zeros:
    digits = digits.lstrip('0')

    # remove trailing zeros:
    if remove_trailing_zeros:
        digits = digits.rstrip('0')

    return len(digits)


def add_one_to_last_digit(num_str):
    num_list = list(num_str)  # Convert string to list for easy modification
    carry = 1
    for i in range(len(num_list) - 1, -1, -1):
        if num_list[i].isdigit():
            if num_list[i] == '9' and carry == 1:
                num_list[i] = '0'  # Carry the increment over to the next digit
            else:
                num_list[i] = str(int(num_list[i]) + carry)  # Add the increment
                carry = 0  # Reset the carry
        if carry == 0:  # No need to go further if there's no carry
            break
    if carry == 1:  # If there's still a carry, prepend it to the list
        num_list = ['1'] + num_list
    return ''.join(num_list)


def split_number_and_power(str_number: str) -> Tuple[str, int]:
    if 'e' in str_number:
        str_number, power = str_number.split('e')
        power = int(power)
    else:
        power = 0
    return str_number, power


def is_number_legit(str_number: str,
                    ignore_int_below: int = 0,
                    ignore_one_with_zeros: bool = True,
                    special_numbers_to_ignore: List[str] = ('95', '99', '100', '1.96', '0.05'),
                    ) -> bool:
    """
    Check if the given string number is ok even if not found in the source.
    """
    if ignore_int_below and is_int_below_max(str_number, ignore_int_below):
        return True

    # we do not check numbers like 1, 0.1, 0.01, etc., or 1, 10, 100, etc.:
    if ignore_one_with_zeros and is_one_with_zeros(str_number):
        return True

    # check if the string number is a special number that we want to ignore:
    if str_number.strip('-').strip('+') in special_numbers_to_ignore:
        return True


def find_non_matching_numeric_values(source: str, target: str, ignore_int_below: int = 0,
                                     remove_trailing_zeros: bool = False,
                                     ignore_one_with_zeros: bool = True,
                                     ignore_after_smaller_than_sign: bool = True,
                                     special_numbers_to_ignore: List[str] = ('95', '99', '100', '1.96', '0.05'),
                                     allow_truncating: bool = True) -> Tuple[List[str], List[str]]:
    """
    Check that all the numerical values mentioned in the target are also mentioned in the source.
    For each numerical value in the target, we check that there exists a numeric values in the source
    that matches after rounding to the same number of digits.
    """

    target = unify_representation_of_numeric_values(target)
    source = unify_representation_of_numeric_values(source)

    str_target_numbers = extract_numeric_values(target)
    str_source_numbers = extract_numeric_values(source)

    non_matching_str_numbers = []
    matching_str_numbers = []
    for str_target_number in str_target_numbers:
        original_str_target_number = str_target_number

        str_target_number = str_target_number.lower()

        if is_number_legit(str_target_number, ignore_int_below, ignore_one_with_zeros, special_numbers_to_ignore):
            continue

        # check if the string number appears after a '<' sign:
        if ignore_after_smaller_than_sign and is_after_smaller_than_sign(str_target_number, target):
            continue

        str_target_number, power = split_number_and_power(str_target_number)

        num_digits = get_number_of_significant_figures(str_target_number, remove_trailing_zeros)

        for should_truncate in range(allow_truncating + 1):

            target_number = round_to_n_digits(str_target_number, num_digits) * 10 ** power
            target_number_if_percent = round(target_number / 100, 10)  # round is just for python float precision

            # check that there exists a number in the source that matches after rounding to the same number of digits:
            if should_truncate:
                is_match_as_is = is_any_matching_value_after_truncating_to_n_digits(
                    str_source_numbers, target_number, num_digits)
                is_match_100 = is_any_matching_value_after_truncating_to_n_digits(
                    str_source_numbers, target_number_if_percent, num_digits)
            else:
                is_match_as_is = is_any_matching_value_after_rounding_to_n_digits(
                    str_source_numbers, target_number, num_digits)
                is_match_100 = is_any_matching_value_after_rounding_to_n_digits(
                    str_source_numbers, target_number_if_percent, num_digits)

            # for now, we assume that any number might be a percentage, setting to None:
            is_target_percentage = None  # is_percentage(str_target_number, target)

            if is_target_percentage is None:  # maybe percentage
                is_match = is_match_as_is or is_match_100
            elif is_target_percentage:  # percentage
                is_match = is_match_100
            else:  # not percentage
                is_match = is_match_as_is
            if is_match:
                matching_str_numbers.append(original_str_target_number)
                break
        else:
            non_matching_str_numbers.append(original_str_target_number)

    return non_matching_str_numbers, matching_str_numbers



from data_to_paper.base_cast import Agent
from data_to_paper.base_cast.types import Profile

from .env import THEME_NAME

# load theme:
theme = THEME_NAME


class DemoAgent(Agent):
    Performer = 'Performer'
    Director = 'Director'
    Debugger = 'Debugger'
    Writer = 'Writer'

    @classmethod
    def get_primary_agent(cls) -> Agent:
        return cls.Performer

    @classmethod
    def get_human_agent(cls) -> Agent:
        return cls.Director

    def get_conversation_name(self) -> str:
        return AGENTS_TO_CONVERSATION_NAMES[self]

    @property
    def profile(self) -> Profile:
        return getattr(theme, self.name)


assert all(agent.profile.agent_name == agent.name for agent in Agent), \
    f"Agent name in theme {THEME_NAME} does not match Agent enum"


AGENTS_TO_CONVERSATION_NAMES = {
    DemoAgent.Performer: None,
    DemoAgent.Director: 'get data',
    DemoAgent.Debugger: 'debug',
    DemoAgent.Writer: 'write paper',
}



import numbers
from contextlib import contextmanager
from functools import partial
from typing import Callable, Any

import numpy as np
import pandas as pd
from pandas import DataFrame

from data_to_paper.env import NUM_DIGITS_FOR_FLOATS
from data_to_paper.run_gpt_code.overrides.dataframes.df_methods.to_latex import to_latex_with_escape
from data_to_paper.run_gpt_code.overrides.dataframes.original_methods import get_original_df_method


def format_numeric_value(value: numbers.Number, float_format: str = None) -> str:
    float_format = float_format or f'.{NUM_DIGITS_FOR_FLOATS}g'

    if isinstance(value, int) or value.is_integer():
        return str(int(value))
    else:
        return f'{value:{float_format}}'


def format_numerics_and_iterables(value: Any, numeric_formater: Callable = None, object_formatter: Callable = None):
    numeric_formater = numeric_formater or format_numeric_value
    object_formatter = object_formatter or (lambda x: str(x))
    if isinstance(value, numbers.Number):
        return numeric_formater(value)
    if isinstance(value, tuple):
        return '(' + ', '.join(format_numerics_and_iterables(v, numeric_formater) for v in value) + ')'
    elif isinstance(value, list):
        return '[' + ', '.join(format_numerics_and_iterables(v, numeric_formater) for v in value) + ']'
    elif isinstance(value, dict):
        return '{' + ', '.join(f'{k}: {format_numerics_and_iterables(v, numeric_formater)}'
                               for k, v in value.items()) + '}'
    elif isinstance(value, np.ndarray):
        return '[' + ', '.join(format_numerics_and_iterables(v, numeric_formater) for v in value) + ']'
    return object_formatter(value)


@contextmanager
def temporarily_change_float_format(new_format):
    """
    Context manager that temporarily changes the float format to the given format.
    """
    original_float_format = pd.get_option('display.float_format')
    try:
        pd.set_option(f'display.float_format', new_format)
        yield
    finally:
        pd.set_option(f'display.float_format', original_float_format)


def _get_formatters_for_df(df: DataFrame, numeric_formater: Callable = None, object_formatter: Callable = None):
    formater = partial(format_numerics_and_iterables, numeric_formater=numeric_formater,
                       object_formatter=object_formatter)
    return [formater] * df.shape[1]


def df_to_string_with_format_value(df: DataFrame, *args,
                                   numeric_formater: Callable = None,
                                   object_formatter: Callable = None, **kwargs):
    to_string = get_original_df_method('to_string')
    return to_string(df, *args, formatters=_get_formatters_for_df(df, numeric_formater, object_formatter), **kwargs)


def df_to_latex_with_value_format(df: DataFrame, *args,
                                  numeric_formater: Callable = None,
                                  object_formatter: Callable = None, **kwargs):
    to_latex = get_original_df_method('to_latex')
    return to_latex_with_escape(df, *args, original_method=to_latex,
                                formatters=_get_formatters_for_df(df, numeric_formater, object_formatter), **kwargs)


def df_to_html_with_value_format(df: DataFrame, *args,
                                 numeric_formater: Callable = None,
                                 object_formatter: Callable = None, **kwargs):
    to_html = get_original_df_method('to_html')
    return to_html(df, *args, formatters=_get_formatters_for_df(df, numeric_formater, object_formatter), **kwargs)


def df_to_llm_readable_csv(df, index_label='index',
                           numeric_formater: Callable = None,
                           object_formatter: Callable = None):
    """
    Convert a DataFrame to a CSV string that is easy to parse by the LLM.
    Takes care of:
    - Quoting string values
    - Quoting column names
    - Quoting index values
    - Formatting lists, tuples, and dictionaries
    - Specifying the index label
    - Multi-level column headers
    - Multi-level indices
    """
    output = []

    # Handle multi-level column headers
    if isinstance(df.columns, pd.MultiIndex):
        for level in range(df.columns.nlevels):
            if index_label and df.index.nlevels > 0:  # Add blank spaces for the index label in header rows
                level_header = [index_label] * df.index.nlevels
            else:
                level_header = []
            level_header += ['"{}"'.format(col) for col in df.columns.get_level_values(level)]
            output.append(','.join(level_header))
    else:
        # Single level columns
        header = ['"{}"'.format(col) for col in df.columns]
        if index_label and df.index.nlevels > 0:
            header = [index_label] * df.index.nlevels + header
        output.append(','.join(header))

    # Handle multi-level indices and quote string indices
    for index, row in df.iterrows():
        index_part = list(index) if isinstance(index, tuple) else [index]
        # Quote strings in index parts
        formatted_index = [f'"{item}"' if isinstance(item, str) else str(item) for item in index_part]
        formatted_row = formatted_index

        for item in row:
            formatted_row.append(str(format_numerics_and_iterables(item, numeric_formater, object_formatter)))
        output.append(','.join(formatted_row))

    return '\n'.join(output)



from data_to_paper.conversation.stage import Stage


class ScientificStage(Stage):
    DATA = ("Get Data", False)
    EXPLORATION = ("Data Exploration", True)
    GOAL = ("Research Goal", True)
    LITERATURE_REVIEW_GOAL = ("Lit. Review I", False)
    ASSESS_NOVELTY = ("Assess Novelty", False)
    PLAN = ("Hypothesis & Plan", True)
    CODE = ("Data Analysis", True)
    DISPLAYITEMS = ("Tables/Figures", True)
    INTERPRETATION = ("Draft abstract", True)
    LITERATURE_REVIEW_WRITING = ("Lit. Review II", True)
    WRITING_RESULTS = ("Results", True)
    WRITING_TITLE_AND_ABSTRACT = ("Title and Abstract", True)
    WRITING_METHODS = ("Methods", True)
    WRITING_INTRODUCTION = ("Introduction", True)
    WRITING_DISCUSSION = ("Discussion", True)
    COMPILE = ("Compile Paper", False)


SECTION_NAMES_TO_WRITING_STAGES = {
    "title_and_abstract": ScientificStage.WRITING_TITLE_AND_ABSTRACT,
    "introduction": ScientificStage.WRITING_INTRODUCTION,
    "methods": ScientificStage.WRITING_METHODS,
    "results": ScientificStage.WRITING_RESULTS,
    "discussion": ScientificStage.WRITING_DISCUSSION,
}



from abc import abstractmethod, ABCMeta
from dataclasses import is_dataclass, fields


class data_to_paperException(Exception, metaclass=ABCMeta):
    """
    Base class for all exceptions in this package.
    """
    @abstractmethod
    def __str__(self):
        pass

    def __reduce__(self):
        if is_dataclass(self):
            # Collect the field values in the order they're defined in the dataclass
            field_values = [getattr(self, f.name) for f in fields(self)]
            return (self.__class__, tuple(field_values))
        else:
            # Fallback for non-dataclass exceptions
            return super().__reduce__()



from __future__ import annotations
import sys
from typing import Optional

from PySide6.QtWidgets import QApplication

from data_to_paper.env import CHOSEN_APP
from .base_app import BaseApp

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from data_to_paper.base_steps import BaseStepsRunner

IS_APP_INITIALIZED = False
THE_APP: Optional[BaseApp] = None
Q_APPLICATION: Optional[QApplication] = None


def get_or_create_q_application_if_app_is_pyside() -> Optional[QApplication]:
    if CHOSEN_APP != 'pyside':
        return None
    global Q_APPLICATION
    if Q_APPLICATION is None:
        Q_APPLICATION = QApplication(sys.argv)
    return Q_APPLICATION


def get_app() -> Optional[BaseApp]:
    global THE_APP, IS_APP_INITIALIZED
    if not IS_APP_INITIALIZED and THE_APP is not None:
        raise ValueError("App is not initialized")
    return THE_APP


def create_app(step_runner: BaseStepsRunner = None) -> Optional[BaseApp]:
    global IS_APP_INITIALIZED, THE_APP
    if IS_APP_INITIALIZED:
        raise ValueError("App is already initialized")

    IS_APP_INITIALIZED = True

    if CHOSEN_APP == None:  # noqa  (Mutable)
        THE_APP = None
        return THE_APP
    if CHOSEN_APP == 'pyside':
        from .pyside_app import PysideApp
        get_or_create_q_application_if_app_is_pyside()
        THE_APP = PysideApp.get_instance()
    elif CHOSEN_APP == 'console':
        from .base_app import ConsoleApp
        THE_APP = ConsoleApp.get_instance()
    else:
        raise ValueError(f"Unknown app type: {CHOSEN_APP}")
    THE_APP.step_runner = step_runner
    return THE_APP



from dataclasses import dataclass

from data_to_paper.utils.serialize import SerializableValue


@dataclass
class HumanAction(SerializableValue):
    """
    Class to store human actions.
    """
    pass


@dataclass
class ButtonClickedHumanAction(HumanAction):
    """
    Class to store human actions of clicking a button.
    value: str - the name of the button.
    """
    pass


@dataclass
class TextSentHumanAction(HumanAction):
    """
    Class to store human actions of sending text.
    value: str - the text sent.
    """
    pass


@dataclass
class RequestInfoHumanAction(HumanAction):
    """
    Class to store human actions of requesting missing text.
    """
    pass



from dataclasses import dataclass
from typing import Dict

from data_to_paper.base_steps import BaseLatexToPDF

from .products import DemoProducts


@dataclass
class ProduceDemoPaperPDF(BaseLatexToPDF):
    products: DemoProducts = None

    def _get_sections(self) -> Dict[str, str]:
        return {section_name: self.products.paper_sections[section_name]
                for section_name in self.paper_section_names}

    def _get_appendix(self):
        s = self.products.data_file_descriptions.to_latex()
        s += '\n\n' + self.products.code_and_output.as_latex_for_appendix()
        return s



from dataclasses import dataclass
from typing import Dict, List, Tuple

from data_to_paper.base_steps import BaseLatexToPDF
from data_to_paper.code_and_output_files.file_view_params import ViewPurpose
from data_to_paper.latex.latex_to_pdf import evaluate_latex_num_command
from data_to_paper.research_types.hypothesis_testing.scientific_products import ScientificProducts
from data_to_paper.code_and_output_files.ref_numeric_values import ReferencedValue, HypertargetFormat, \
    HypertargetPosition
from data_to_paper.code_and_output_files.referencable_text import ListReferenceableText
from data_to_paper.servers.custom_types import Citation


@dataclass
class ProduceScientificPaperPDFWithAppendix(BaseLatexToPDF):
    products: ScientificProducts = None

    def _get_formatted_section_and_notes(self, section_name: str) -> Tuple[str, Dict[str, str]]:
        section = self.products.get_tabled_paper_sections(ViewPurpose.FINAL_INLINE)[section_name]
        return evaluate_latex_num_command(section, ref_prefix=section_name.replace(' ', '_'))

    def _get_sections(self) -> Dict[str, str]:
        return {section_name: self._get_formatted_section_and_notes(section_name)[0]
                for section_name in self.paper_section_names}

    def _get_references(self) -> List[Citation]:
        return self.products.get_all_cited_citations()

    def _get_all_notes(self):
        notes = {}
        for section_name in self.paper_section_names:
            _, section_notes = self._get_formatted_section_and_notes(section_name)
            notes.update(section_notes)
        return notes

    def _get_notes_appendix(self):
        notes = self._get_all_notes()
        if not notes:
            return ''
        template = '\n'.join(r'\item{@' + value + r'}' for key, value in notes.items())
        text = ListReferenceableText(text=template,
                                     hypertarget_prefix='',
                                     pattern='@',
                                     reference_list=[ReferencedValue(label=key, value='', is_target=True)
                                                     for key, value in notes.items()]
                                     )

        # no_indent = "\\setlength{\\itemindent}{0em}\n\\setlength{\\leftmargini}{0em}\n"
        no_indent = ""
        return f"\\section{{Calculation Notes}}\n{no_indent}\\begin{{itemize}}\n" \
            + text.get_formatted_text(hypertarget_format=HypertargetFormat(position=HypertargetPosition.WRAP)) \
            + "\n\\end{itemize}"

    def _get_appendix(self):
        s = ''
        s += self.products.data_file_descriptions.to_latex(view_purpose=ViewPurpose.FINAL_APPENDIX)
        for code_name, code_and_output in self.products.codes_and_outputs.items():
            s += '\n\n' + code_and_output.as_latex_for_appendix()
        notes_appendix = self._get_notes_appendix()
        if notes_appendix:
            s += '\n\n' + notes_appendix
        return s



import json
from dataclasses import dataclass

from data_to_paper import Message
from data_to_paper.base_steps.base_products_conversers import ReviewBackgroundProductsConverser

from typing import Any, Dict, Optional, get_origin, Collection, Iterable

from data_to_paper.base_steps.result_converser import Rewind
from data_to_paper.env import JSON_MODEL_ENGINE
from data_to_paper.run_gpt_code.code_utils import extract_content_of_triple_quote_block, FailedExtractingBlock, \
    NoBlocksFailedExtractingBlock, IncompleteBlockFailedExtractingBlock
from data_to_paper.servers.model_engine import ModelEngine
from data_to_paper.utils.nice_list import NiceDict
from data_to_paper.utils.tag_pairs import TagPairs
from data_to_paper.utils.check_type import validate_value_type, WrongTypeException
from data_to_paper.text.text_extractors import extract_text_between_most_flanking_tags
from data_to_paper.text import wrap_as_block

TYPES_TO_TAG_PAIRS: Dict[type, TagPairs] = {
    dict: TagPairs('{', '}'),
    list: TagPairs('[', ']'),
    tuple: TagPairs('(', ')'),
    set: TagPairs('{', '}'),
}


@dataclass
class PythonValueReviewBackgroundProductsConverser(ReviewBackgroundProductsConverser):
    """
    A base class for agents requesting the LLM to write a python value (like a list of str, or dict).
    Option for reviewing the sections (set max_reviewing_rounds > 0).
    """
    model_engine: ModelEngine = JSON_MODEL_ENGINE
    value_type: type = None
    rewind_after_getting_a_valid_response: Optional[Rewind] = Rewind.AS_FRESH
    json_mode: bool = False
    your_response_should_be_formatted_as: str = '{property_your_response_should_be_formatted_as}'

    @property
    def is_really_json_mode(self) -> bool:
        return self.json_mode and self.model_engine.allows_json_mode

    @property
    def python_or_json(self) -> str:
        return 'json' if self.json_mode else 'Python'

    @property
    def property_your_response_should_be_formatted_as(self) -> str:
        if self.json_mode:
            return f"a json object that can be evaluated with `json.loads()` to a Python {self.type_name}."
        return f"a Python {self.type_name} wrapped within a triple backtick 'python' code block."

    @property
    def parent_type(self) -> type:
        return get_origin(self.value_type)

    @property
    def type_name(self) -> str:
        return str(self.value_type).replace('typing.', '')

    def apply_get_and_append_assistant_message(self, *args, **kwargs) -> Message:
        kwargs['is_json'] = self.is_really_json_mode
        return super().apply_get_and_append_assistant_message(*args, **kwargs)

    def get_valid_result_as_markdown(self) -> str:
        return wrap_as_block(self.valid_result, 'python')

    def _check_response_and_get_extracted_text(self, response: str) -> str:
        """
        Extracts the string of the python value from LLM response.
        If there is an error extracting the value, _raise_self_response_error is called.
        """
        if self.is_really_json_mode:
            return response

        try:
            return extract_content_of_triple_quote_block(response, self.goal_noun, self.python_or_json)
        except NoBlocksFailedExtractingBlock:
            pass
        except FailedExtractingBlock as e:
            self._raise_self_response_error(
                title='# Failed to extract python block',
                error_message=str(e),
                missing_end=isinstance(e, IncompleteBlockFailedExtractingBlock))

        tags = TYPES_TO_TAG_PAIRS.get(self.parent_type)
        try:
            return extract_text_between_most_flanking_tags(response, *tags, keep_tags=True)
        except ValueError:
            self._raise_self_response_error(
                title='# Incorrect response format',
                error_message=f'Could not find a valid Python {self.type_name} in your response.',
                missing_end=tags[0] in response and tags[1] not in response)

    def _check_extracted_text_and_update_valid_result(self, extracted_text: str):
        response_value = self._evaluate_python_value_from_str(extracted_text)
        response_value = self._validate_value_type(response_value)
        response_value = self._check_response_value(response_value)
        self._update_valid_result(response_value)

    def _convert_extracted_text_to_fresh_looking_response(self, extracted_text: str) -> str:
        """
        Return a response that contains just the python value.
        """
        if self.is_really_json_mode:
            return extracted_text
        return wrap_as_block(extracted_text, self.python_or_json)

    def _evaluate_python_value_from_str(self, response: str) -> Any:
        try:
            if self.json_mode:
                return json.loads(response)
            else:
                return eval(response)
        except Exception as e:
            func = 'json.loads()' if self.json_mode else 'eval()'
            formatting_instructions = "{formatting_instructions_for_feedback}" + \
                f"I need to be able to just cut and paste it and evaluate with `{func}`."
            if not self.json_mode:
                formatting_instructions += "\nSo it has to be a valid Python value (not an assignment statement)."
            self._raise_self_response_error(
                title='# Incorrect response format',
                error_message=f'I tried to eval your response with Python `{func}`, but got:\n{e}\n',
                formatting_instructions=formatting_instructions
            )

    def _validate_value_type(self, response_value: Any) -> Any:
        """
        Validate that the response is given in the correct format. if not raise TypeError.
        """
        try:
            validate_value_type(response_value, self.value_type)
        except WrongTypeException as e:
            self._raise_self_response_error(
                title='# Incorrect response type',
                error_message=str(e))
        return response_value

    def _check_response_value(self, response_value: Any) -> Any:
        """
        Check that the response value is valid.
        Return the value if it is valid, otherwise call self._raise_self_response_error.
        The returned value can also be modified.
        """
        return response_value


@dataclass
class PythonDictReviewBackgroundProductsConverser(PythonValueReviewBackgroundProductsConverser):
    """
    A base class for agents requesting the LLM to write a python dict.
    """
    value_type: type = Dict[Any, Any]

    def _check_response_value(self, response_value: Any) -> Any:
        value = super()._check_response_value(response_value)
        return NiceDict(value)


@dataclass
class PythonDictWithDefinedKeysReviewBackgroundProductsConverser(PythonDictReviewBackgroundProductsConverser):
    """
    A base class for agents requesting the LLM to write a python dict, with specified keys.
    """
    requested_keys: Collection[str] = None  # The keys that the dict should contain. `None` means any keys are allowed.
    value_type: type = Dict[str, Any]

    def _check_response_value(self, response_value: Any) -> Any:
        """
        Check that the response value is valid.
        raise a feedback message if it is not valid.
        """
        check_response_value = super()._check_response_value(response_value)
        if self.requested_keys is not None:
            if set(response_value.keys()) != set(self.requested_keys):
                self._raise_self_response_error(
                    title='# Incorrect keys in response',
                    error_message=f'Your response should include a {self.python_or_json} dict containing the keys: '
                                  f'{self.requested_keys}')

        return check_response_value


@dataclass
class PythonDictWithDefinedKeysAndValuesReviewBackgroundProductsConverser(
        PythonDictWithDefinedKeysReviewBackgroundProductsConverser):
    allowed_values_for_keys: Dict[str, Iterable] = None  # The values that the dict may contain.

    def __post_init__(self):
        super().__post_init__()
        assert self.requested_keys is None
        self.requested_keys = self.allowed_values_for_keys.keys()

    def _check_response_value(self, response_value):
        check_response_value = super()._check_response_value(response_value)
        for key, value in response_value.items():
            if value not in self.allowed_values_for_keys[key]:
                self._raise_self_response_error(
                    title='# Incorrect values in python dict',
                    error_message=f'The value for key `{key}` should be one of: {self.allowed_values_for_keys[key]}')
        return check_response_value



from __future__ import annotations

import importlib
import pkgutil
from dataclasses import dataclass

import inspect

from typing import Callable, Iterable, Any, Optional, Tuple, Dict, Union

from .base_run_contexts import RegisteredRunContext
from .exceptions import CodeUsesForbiddenFunctions
from .run_issues import CodeProblem, RunIssue


def _carefully_get_members(module):
    try:
        return inspect.getmembers(module)
    except Exception:
        return []


def get_all_submodules(module, visited=None):
    """Recursively get all sub-modules of a module, avoiding loops."""
    if visited is None:
        visited = set()

    all_submodules = []

    for name, obj in _carefully_get_members(module):
        if inspect.ismodule(obj) and obj not in visited:
            visited.add(obj)
            all_submodules.append(obj)
            all_submodules.extend(get_all_submodules(obj, visited))

    return all_submodules


def import_submodules(package):
    """ Recursively import all submodules of a module, including sub-packages """
    if isinstance(package, str):
        package = importlib.import_module(package)
    results = {}
    for loader, name, is_pkg in pkgutil.walk_packages(package.__path__):
        full_name = package.__name__ + '.' + name
        try:
            results[full_name] = importlib.import_module(full_name)
            if is_pkg:
                results.update(import_submodules(full_name))
        except ModuleNotFoundError:
            continue
    return results


def dynamic_import(full_path):
    """Dynamically import a module or class from a string path."""
    try:
        # This will work if the path is to a module (e.g., 'statsmodels.stats.multitest')
        return importlib.import_module(full_path)
    except ImportError:
        # If direct import fails, it might be a class or function in a module
        if '.' not in full_path:
            raise
        module_path, class_or_func_name = full_path.rsplit('.', 1)
        module = importlib.import_module(module_path)
        return getattr(module, class_or_func_name)


def _import_obj(obj_import_str: Optional[str, Any]):
    if obj_import_str is None:
        return None
    if isinstance(obj_import_str, str):
        return dynamic_import(obj_import_str)
    return obj_import_str


@dataclass
class OverrideImportedObjContext(RegisteredRunContext):
    obj_import_str: Optional[str, Any] = None

    @property
    def obj(self):
        return _import_obj(self.obj_import_str)


@dataclass
class MultiAttrReplacerContext(OverrideImportedObjContext):
    _originals: Optional[dict] = None

    def _get_all_parents(self) -> set:
        raise NotImplementedError

    def _get_all_attrs_for_parent(self, parent) -> Iterable[str]:
        raise NotImplementedError

    def _get_custom_wrapper(self, parent, attr_name, original_func):
        raise NotImplementedError

    def __enter__(self):
        self._originals = {}
        for parent in self._get_all_parents():
            for attr_name in self._get_all_attrs_for_parent(parent):
                original = getattr(parent, attr_name)
                self._originals[(parent, attr_name)] = original
                setattr(parent, attr_name, self._get_custom_wrapper(parent, attr_name, original))
        return super().__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        for (parent, attr_name), original in self._originals.items():
            setattr(parent, attr_name, original)
        self._originals = None
        return super().__exit__(exc_type, exc_val, exc_tb)


@dataclass
class SystematicAttrReplacerContext(MultiAttrReplacerContext):
    recursive: bool = True

    def _get_all_modules(self) -> list:
        all_modules = [self.obj]
        if self.recursive:
            all_modules += get_all_submodules(self.obj)
        return all_modules

    def _is_right_type(self, obj) -> bool:
        raise NotImplementedError

    def _should_replace(self, parent, attr_name, attr) -> bool:
        raise NotImplementedError

    def _get_all_attrs_for_parent(self, parent) -> Iterable[str]:
        return [attr_name for attr_name, attr_obj in parent.__dict__.items()
                if self._is_right_type(attr_obj) and self._should_replace(parent, attr_name, attr_obj)]


class SystematicMethodReplacerContext(SystematicAttrReplacerContext):
    def _get_all_parents(self) -> list:
        classes = []
        for mod in self._get_all_modules():
            for name, obj in _carefully_get_members(mod):
                if inspect.isclass(obj) and obj not in classes:
                    classes.append(obj)
        return classes

    def _is_right_type(self, obj) -> bool:
        return inspect.isfunction(obj) or inspect.ismethod(obj)


class SystematicFuncReplacerContext(SystematicAttrReplacerContext):
    def _get_all_parents(self) -> list:
        return self._get_all_modules()

    def _is_right_type(self, obj) -> bool:
        return inspect.isfunction(obj)


@dataclass
class AttrReplacer(OverrideImportedObjContext):
    attr: str = None
    wrapper: Union[Callable, Any] = None

    send_context_to_wrapper: bool = False
    send_original_to_wrapper: bool = False

    _original: Callable = None

    def _get_wrapper(self):
        return self.wrapper

    def _get_wrapped_wrapper(self):
        if not callable(self._get_wrapper()):
            return self._get_wrapper()

        def wrapped_wrapper(*args, **kwargs):
            if not self._is_enabled or not self._is_called_from_user_script():
                return self._original(*args, **kwargs)
            if self.send_context_to_wrapper:
                kwargs['context_manager'] = self
            if self.send_original_to_wrapper:
                kwargs['original_func'] = self._original
            return self._get_wrapper()(*args, **kwargs)
        return wrapped_wrapper

    def _reversible_enter(self):
        self._original = getattr(self.obj, self.attr)
        setattr(self.obj, self.attr, self._get_wrapped_wrapper())
        super()._reversible_enter()

    def _reversible_exit(self):
        setattr(self.obj, self.attr, self._original)
        super()._reversible_exit()


@dataclass
class PreventAssignmentToAttrs(OverrideImportedObjContext):
    forbidden_set_attrs: Iterable[str] = ()
    message: str = 'Cannot set {attr}.'
    _original: Callable = None

    def _get_wrapper(self):
        def _replacement__setattr__(obj, attr, value):
            if not self._is_enabled or not self._is_called_from_user_script():
                return self._original(obj, attr, value)
            if attr in self.forbidden_set_attrs:
                self._raise_exception(attr, value)
            return self._original(obj, attr, value)
        return _replacement__setattr__

    def _raise_exception(self, attr, value):
        raise AttributeError(self.message.format(attr=attr))

    def _reversible_enter(self):
        self._original = self.obj.__setattr__
        self.obj.__setattr__ = self._get_wrapper()
        super()._reversible_enter()

    def _reversible_exit(self):
        self.obj.__setattr__ = self._original
        super()._reversible_exit()


@dataclass
class PreventCalling(RegisteredRunContext):
    modules_and_functions: Iterable[Tuple[Any, str, bool]] = None
    _original_functions: Optional[Dict[str, Callable]] = None

    def __enter__(self):
        self._original_functions = {}
        for module, function_name, should_only_create_issue in self.modules_and_functions:
            module = _import_obj(module)
            original_func = getattr(module, function_name)
            setattr(module, function_name, self.get_upon_called(function_name, original_func, should_only_create_issue))
            self._original_functions[function_name] = original_func
        return super().__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        for module, function_name, _ in self.modules_and_functions:
            module = _import_obj(module)
            setattr(module, function_name, self._original_functions.pop(function_name))
        self._original_functions = None
        return super().__exit__(exc_type, exc_val, exc_tb)

    def get_upon_called(self, func_name: str, original_func: Callable, should_only_create_issue: bool):
        def upon_called(*args, **kwargs):
            if not self._is_enabled or not self._is_called_from_user_script():
                return original_func(*args, **kwargs)
            if should_only_create_issue:
                self.issues.append(RunIssue.from_current_tb(
                    category='Coding: good practices',
                    issue=f'Code uses forbidden function: "{func_name}".',
                    code_problem=CodeProblem.NonBreakingRuntimeIssue,
                ))
            else:
                raise CodeUsesForbiddenFunctions(func_name)
            return original_func(*args, **kwargs)
        return upon_called



from dataclasses import dataclass
from pathlib import Path
from typing import Optional, Tuple

from data_to_paper.text import dedent_triple_quote_str
from data_to_paper.utils.nice_list import NiceList

from data_to_paper.base_steps import BaseCodeProductsGPT
from data_to_paper.code_and_output_files.output_file_requirements import TextContentOutputFileRequirement, \
    OutputFileRequirements
from data_to_paper.servers.model_engine import ModelEngine

from .cast import DemoAgent
from .products import DemoProducts


@dataclass
class DemoCodeProductsGPT(BaseCodeProductsGPT):
    model_engine: ModelEngine = ModelEngine.GPT4_TURBO
    products: DemoProducts = None
    assistant_agent: DemoAgent = DemoAgent.Performer
    user_agent: DemoAgent = DemoAgent.Debugger
    background_product_fields: Tuple[str, ...] = ('data_file_descriptions', 'research_goal')
    code_name: str = 'Prime Number Search'

    @property
    def data_filenames(self) -> NiceList[str]:
        return NiceList(self.products.data_file_descriptions.get_data_filenames(),
                        wrap_with='"',
                        prefix='{} data file[s]: ')

    @property
    def data_folder(self) -> Optional[Path]:
        return Path(self.products.data_file_descriptions.data_folder)

    supported_packages: Tuple[str, ...] = ('numpy', )

    mission_prompt: str = dedent_triple_quote_str("""
        Please write a short Python code for the goal above.

        Your code should create the following output files: {requested_output_filenames}.
        These files should contain ...

        If needed, you can use the following packages which are already installed:
        {supported_packages}
        """)

    def _create_output_file_requirements(self) -> OutputFileRequirements:
        return OutputFileRequirements([TextContentOutputFileRequirement('output.txt')])



"""
This script allows manually running a GPT-generated code snippet in the correct context.

Paste the code into the `gpt_code` function.
Specify:
- the `code_converser` that created the GPT code.
- the `run_folder` where the code will be run (where the data is stored).
    If not provided, the code will run in the default run_folder (FOLDER_FOR_RUN).
"""
from data_to_paper.env import FOLDER_FOR_RUN
from data_to_paper.research_types.hypothesis_testing.coding.analysis import DataAnalysisCodeProductsGPT
from data_to_paper.run_gpt_code.run_code_in_context import run_code_in_context

# specify the converser that created the gpt code:
code_converser_cls = DataAnalysisCodeProductsGPT

# specify the run folder where the code will be run:
run_folder = FOLDER_FOR_RUN  # will run in the 'temp_run' folder


def gpt_code():
    # COPY HERE THE GPT CODE
    pass


if __name__ == '__main__':
    code_converser = code_converser_cls()
    results = run_code_in_context(gpt_code,
                                  code_converser=code_converser,
                                  run_folder=run_folder)
    _, _, _, exception = results
    print(exception)



from dataclasses import dataclass
from typing import Any

from data_to_paper.code_and_output_files.file_view_params import ViewPurpose
from data_to_paper.conversation.stage import Stage
from data_to_paper.text.highlighted_text import format_text_with_code_blocks


@dataclass
class Product:
    """
    A product is a piece of information that is generated during the conversation.
    A product needs to know how to be presented:
    - as markdown, to be included in a pre-conversation context.
    - as html, to be shown in the app.
    """
    name: str = None
    stage: Stage = None

    def get_stage(self, **kwargs) -> Stage:
        return self.stage

    def is_valid(self):
        raise NotImplementedError

    def _get_content_as_formatted_text(self, level: int, view_purpose: ViewPurpose, **kwargs) -> str:
        raise NotImplementedError

    def _get_content_as_html(self, level: int, **kwargs) -> str:
        return format_text_with_code_blocks(self._get_content_as_formatted_text(level, ViewPurpose.APP_HTML, **kwargs),
                                            from_md=True, is_html=True, width=None)

    def get_header(self, view_purpose: ViewPurpose = ViewPurpose.PRODUCT, **kwargs) -> str:
        return self.name

    def as_formatted_text(self, level: int = 1, with_header: bool = True,
                          view_purpose: ViewPurpose = ViewPurpose.PRODUCT,
                          **kwargs) -> str:
        """
        Return the product in the form to be included in a pre-conversation context.
        Typically, this is a markdown format.
        """
        s = ''
        if with_header:
            s += '#' * level + ' ' + self.get_header(**kwargs) + '\n'
        s += self._get_content_as_formatted_text(level, view_purpose=view_purpose, **kwargs)
        return s

    def as_html(self, level: int = 0, with_header: bool = True, **kwargs) -> str:
        s = ''
        if with_header:
            s += f'<h{level}>{self.get_header(**kwargs)}</h{level}>'
        s += self._get_content_as_html(level, **kwargs)
        return s


@dataclass
class ValueProduct(Product):
    value: Any = None

    def is_valid(self):
        return self.value is not None

    def _get_content_as_formatted_text(self, level: int, view_purpose: ViewPurpose, **kwargs) -> str:
        return str(self.value)

    def __getitem__(self, item):
        return self.value[item]

    def __setitem__(self, key, value):
        self.value[key] = value

    def __iter__(self):
        return iter(self.value)

    def items(self):
        return self.value.items()

    def keys(self):
        return self.value.keys()

    def values(self):
        return self.value.values()

    def __len__(self):
        return len(self.value)

    def __contains__(self, item):
        return item in self.value



from functools import partial
from typing import Optional, List, Collection, Dict, Callable, Any, Union, Tuple

from PySide6.QtCore import Qt, QMutex, QWaitCondition, QThread, Signal, Slot
from PySide6.QtGui import QTextOption, QTextCursor
from PySide6.QtWidgets import QMainWindow, QVBoxLayout, QLabel, QPushButton, QWidget, \
    QHBoxLayout, QSplitter, QTextEdit, QTabWidget, QDialog, QSizePolicy, QCheckBox, QSpacerItem

from data_to_paper.conversation.stage import Stage
from data_to_paper.interactive.base_app import BaseApp, REQUESTING_MISSING_TEXT
from data_to_paper.interactive.enum_types import PanelNames
from data_to_paper.interactive.get_app import get_or_create_q_application_if_app_is_pyside
from data_to_paper.interactive.styles import CURRENT_STEP_COLOR, SUBMIT_BUTTON_COLOR, PANEL_HEADER_COLOR, QEDIT_STYLE, \
    STEP_PANEL_BUTTON_STYLE, BACKGROUND_COLOR, CSS, APP_STYLE, TABS_STYLE, HTMLPOPUP_STYLE, \
    MAIN_SPLITTER_STYLE, QCHECKBOX_STYLE, STEP_PANEL_RESET_BUTTON_STYLE
from data_to_paper.interactive.utils import open_file_on_os
from data_to_paper.servers.api_cost import StageToCost


def _get_label_height(label: QLabel) -> int:
    """
    Get the height of a one-row QLabel.
    """
    return label.fontMetrics().height() + 2 * label.contentsMargins().top()


class Worker(QThread):
    # Signal now carries a string payload for initial text
    request_panel_continue_signal = Signal(PanelNames)
    request_text_signal = Signal(PanelNames, str, str, str, str, dict)
    show_text_signal = Signal(PanelNames, str, bool, bool)
    set_focus_on_panel_signal = Signal(PanelNames)
    advance_stage_int_signal = Signal(int)
    send_product_of_stage_signal = Signal(Stage, str)
    set_status_signal = Signal(PanelNames, int, str)
    set_header_signal = Signal(str)
    send_api_usage_cost_signal = Signal(object)

    def __init__(self, mutex, condition, func_to_run=None):
        super().__init__()
        self.mutex = mutex
        self.condition = condition
        self.func_to_run = func_to_run
        self._text_input = None
        self._panel_name_input = None

    def run(self):
        if self.func_to_run is not None:
            self.func_to_run()

    def worker_set_status(self, panel_name: PanelNames, position: int, status: str = ''):
        self.set_status_signal.emit(panel_name, position, status)

    def worker_set_header(self, header: str):
        self.set_header_signal.emit(header)

    def worker_request_panel_continue(self, panel_name: PanelNames):
        self.mutex.lock()
        self.request_panel_continue_signal.emit(panel_name)
        self.condition.wait(self.mutex)
        self.mutex.unlock()

    def worker_request_text(self, panel_name: PanelNames, initial_text: str = '',
                            title: Optional[str] = None,
                            instructions: Optional[str] = None,
                            in_field_instructions: Optional[str] = None,
                            optional_suggestions: Dict[str, str] = None) -> str:
        self.mutex.lock()
        self.request_text_signal.emit(panel_name, initial_text, title, instructions, in_field_instructions,
                                      optional_suggestions)
        self.condition.wait(self.mutex)
        input_text = self._text_input
        self.mutex.unlock()
        return input_text

    def worker_show_text(self, panel_name: PanelNames, text: str, is_html: bool = False,
                         scroll_to_bottom: bool = False):
        self.show_text_signal.emit(panel_name, text, is_html, scroll_to_bottom)

    def worker_set_focus_on_panel(self, panel_name: PanelNames):
        self.set_focus_on_panel_signal.emit(panel_name)

    def worker_advance_stage_int(self, stage: int):
        self.advance_stage_int_signal.emit(stage)

    def worker_send_product_of_stage(self, stage: Stage, product_text: str):
        self.send_product_of_stage_signal.emit(stage, product_text)

    def worker_send_api_usage_cost(self, stages_to_costs: Dict[str, float]):
        self.send_api_usage_cost_signal.emit(stages_to_costs)

    @Slot(PanelNames, str)
    def receive_text_signal(self, panel_name, text):
        self.mutex.lock()
        self._text_input = text
        self._panel_name_input = panel_name
        self.condition.wakeAll()
        self.mutex.unlock()

    @Slot()
    def receive_continue_signal(self):
        self.mutex.lock()
        self.condition.wakeAll()
        self.mutex.unlock()

    @Slot(PanelNames)
    def receive_panel_continue_signal(self, panel_name):
        self.mutex.lock()
        self.condition.wakeAll()
        self.mutex.unlock()


class StepsPanel(QWidget):
    def __init__(self):
        super().__init__()
        self.labels_to_callbacks = None
        self.current_step = 0
        self.layout = QVBoxLayout(self)
        self.step_widgets = {}

    def init_ui(self, labels_to_callbacks: Dict[Tuple[str, str, bool], Tuple[Callable, Callable]]):
        self.labels_to_callbacks = labels_to_callbacks
        for (label, name, resettable), (step_callback, reset_callback) in self.labels_to_callbacks.items():
            self.step_widgets[label] = QHBoxLayout()
            step_button = QPushButton(label)
            step_button.setFixedWidth(130)
            step_button.clicked.connect(step_callback)
            self.step_widgets[label].addWidget(step_button)
            if resettable:
                reset_to_step_button = QPushButton('↺')
                reset_to_step_button.setFixedWidth(20)
                reset_to_step_button.setStyleSheet(STEP_PANEL_RESET_BUTTON_STYLE.format(background_color="#909090",
                                                                                        pressed_color="#707070"))
                reset_to_step_button.clicked.connect(reset_callback)
                self.step_widgets[label].addWidget(reset_to_step_button)
            else:  # add a spacer to keep the layout consistent
                spacer = QSpacerItem(20, 20, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum)
                self.step_widgets[label].addItem(spacer)
            self.layout.addLayout(self.step_widgets[label])
        self.layout.setSpacing(5)
        self.refresh()

    def refresh(self):
        for i, step in enumerate(self.step_widgets):
            if i == self.current_step:
                self.step_widgets[step].itemAt(0).widget().setStyleSheet(
                    STEP_PANEL_BUTTON_STYLE.format(background_color=CURRENT_STEP_COLOR,
                                                   pressed_color="#003377"))
                if self.step_widgets[step].itemAt(1).widget() is not None:
                    self.step_widgets[step].itemAt(1).widget().setEnabled(True)
            elif i < self.current_step:
                self.step_widgets[step].itemAt(0).widget().setStyleSheet(
                    STEP_PANEL_BUTTON_STYLE.format(background_color=SUBMIT_BUTTON_COLOR,
                                                   pressed_color="#006400"))
                if self.step_widgets[step].itemAt(1).widget() is not None:
                    self.step_widgets[step].itemAt(1).widget().setEnabled(True)
            else:
                self.step_widgets[step].itemAt(0).widget().setStyleSheet(
                    STEP_PANEL_BUTTON_STYLE.format(background_color="#909090", pressed_color="#707070"))
                # if the there is a reset button and not a spacer, disable the reset button
                if self.step_widgets[step].itemAt(1).widget() is not None:
                    self.step_widgets[step].itemAt(1).widget().setEnabled(False)

    def set_step_by_index(self, step_index: int):
        self.current_step = step_index
        self.refresh()

    def disable_refresh_of_all_steps(self):
        for step in self.step_widgets:
            if self.step_widgets[step].itemAt(1).widget() is not None:
                # set reset button to not visible
                self.step_widgets[step].itemAt(1).widget().setVisible(False)
                # add a spacer to keep the layout consistent
                spacer = QSpacerItem(20, 20, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum)
                self.step_widgets[step].addItem(spacer)


class Panel(QWidget):
    def __init__(self, header: Optional[str] = None, header_right: Optional[str] = None):
        """
        A panel that displays text and allows editing.
        """
        super().__init__()
        self.layout = QVBoxLayout()
        self.setLayout(self.layout)

        header_tray = QHBoxLayout()
        self.layout.addLayout(header_tray)

        self.header = header
        label = QLabel(header)
        label.setFixedHeight(_get_label_height(label))
        self.header_label = label
        self.header_label.setStyleSheet(f"color: {PANEL_HEADER_COLOR}; font-size: 16px; font-weight: bold;")
        header_tray.addWidget(label)

        self.header_right = header_right
        if header_right is not None:
            right_label = QLabel(header_right)
            right_label.setFixedHeight(_get_label_height(right_label))
            right_label.setStyleSheet("color: #cc0000; font-size: 12px;")
            self.header_right_label = right_label
            header_tray.addWidget(right_label, alignment=Qt.AlignmentFlag.AlignRight)
        else:
            self.header_right_label = None

    def set_header(self, header: str):
        self.header = header
        self.header_label.setText(header)

    def set_header_right(self, text: str):
        if self.header_right_label is None:
            return
        self.header_right = text
        self.header_right_label.setText(text)

    def set_text(self, text):
        pass

    def get_text(self):
        pass


class EditableTextPanel(Panel):
    def __init__(self, header: str, header_right: Optional[str] = None,
                 suggestion_button_names: Optional[Collection[str]] = None):
        super().__init__(header, header_right)
        if suggestion_button_names is None:
            suggestion_button_names = []
        self.suggestion_button_names = suggestion_button_names
        self.suggestion_buttons = []
        self.suggestion_texts = [''] * len(suggestion_button_names)

        self.text_edit = QTextEdit()
        self.text_edit.setWordWrapMode(QTextOption.WrapMode.WrapAtWordBoundaryOrAnywhere)
        # self.text_edit.setFontPointSize(14)
        self.text_edit.setStyleSheet(QEDIT_STYLE)

        self.text_edit.setReadOnly(True)
        self.layout.addWidget(self.text_edit)

        # Instructions:
        self.instructions = None
        self.instructions_label = QLabel()
        self.update_instructions()
        self.layout.addWidget(self.instructions_label)

        # Buttons:
        self.buttons_tray = QHBoxLayout()
        self.layout.addLayout(self.buttons_tray)

        self.continue_button = QPushButton("Continue")
        self.continue_button.setVisible(False)
        self.continue_button.setStyleSheet('QPushButton {background-color: #E3E0DA; color:' + BACKGROUND_COLOR + ';}')
        self.continue_button.clicked.connect(self.on_continue)
        self.buttons_tray.addWidget(self.continue_button)

        self.submit_button = QPushButton("Submit")
        self.submit_button.setStyleSheet('QPushButton {background-color: ' + SUBMIT_BUTTON_COLOR + '; color:'
                                         + BACKGROUND_COLOR + ';}')
        self.submit_button.clicked.connect(self.on_submit)
        self.buttons_tray.addWidget(self.submit_button)

        for i, button_text in enumerate(suggestion_button_names):
            button = QPushButton(button_text)
            button.setStyleSheet('QPushButton {background-color: #E3E0DA; color:' + BACKGROUND_COLOR + ';}')
            button.clicked.connect(self.on_suggestion_button_click)
            self.buttons_tray.addWidget(button)
            self.suggestion_buttons.append(button)
        self._set_buttons_visibility(False)

        # self.setStyleSheet("color: white;")

    def _set_buttons_visibility(self, visible: bool):
        self.submit_button.setVisible(visible)
        self.instructions_label.setVisible(visible)
        for button in self.suggestion_buttons:
            button.setVisible(visible)

    def update_instructions(self):
        if self.instructions is not None:
            self.instructions_label.setText(self.instructions)
        else:
            self.instructions_label.setText('')
        self.instructions_label.setVisible(bool(self.instructions))

    def on_suggestion_button_click(self):
        button = self.sender()
        suggestion_index = self.suggestion_buttons.index(button)
        if suggestion_index >= len(self.suggestion_texts):
            return
        suggestion_text = self.suggestion_texts[suggestion_index]
        if suggestion_text is None:
            self.text_edit.setPlainText(REQUESTING_MISSING_TEXT)
            self.submit_button.click()
        else:
            self.text_edit.setPlainText(suggestion_text)

    def _set_plain_text(self, text: str):
        self.text_edit.setPlainText(text)

    def _set_html_text(self, text: str):
        # add the CSS to the HTML
        self.text_edit.setHtml(f'<style>{CSS}</style>{text}')

    def set_text(self, text: str, is_html: bool = False):
        self.text_edit.setReadOnly(True)
        if is_html:
            self._set_html_text(text)
        else:
            self._set_plain_text(text)

    def set_instructions(self, instructions: str):
        self.instructions = instructions
        self.update_instructions()

    def edit_text(self, text: Optional[str] = '',
                  title: Optional[str] = None,
                  instruction: Optional[str] = None,
                  in_field_instructions: Optional[str] = None,
                  suggestion_texts: Optional[List[str]] = None):
        self.text_edit.setReadOnly(False)
        if in_field_instructions:
            self.text_edit.setPlaceholderText(in_field_instructions)
        self._set_plain_text(text)
        self._set_buttons_visibility(True)
        if suggestion_texts is not None:
            self.suggestion_texts = suggestion_texts
        self.set_instructions(instruction or '')
        self.set_header_right(title or '')

    def scroll_to_bottom(self):
        self.text_edit.moveCursor(QTextCursor.End)
        self.text_edit.ensureCursorVisible()

    def wait_for_continue(self):
        self.continue_button.setVisible(True)

    def on_submit(self):
        self.text_edit.setReadOnly(True)
        self.text_edit.setPlaceholderText('')
        self._set_buttons_visibility(False)
        self.set_header_right('')
        self.set_instructions('')

    def on_continue(self):
        self.continue_button.setVisible(False)

    def get_text(self):
        return self.text_edit.toPlainText()


class HtmlPopup(QDialog):
    def __init__(self, title: str, html_content, parent=None):
        super(HtmlPopup, self).__init__(parent, Qt.WindowType.WindowCloseButtonHint)
        self.setWindowTitle(title)

        # Layout to organize widgets
        layout = QVBoxLayout()

        # QLabel to display HTML content
        label = QTextEdit()
        label.setReadOnly(True)
        label.setHtml(f'<style>{CSS}</style>{html_content}')

        # label.setText(html_content)
        # label.setTextFormat(Qt.RichText)  # Set the text format to RichText to enable HTML
        layout.addWidget(label)

        # QPushButton to close the dialog
        self.close_button = QPushButton("Close")
        self.close_button.clicked.connect(self.close)
        layout.addWidget(self.close_button)

        self.setLayout(layout)
        self.setStyleSheet(HTMLPOPUP_STYLE)
        self.resize(800, 600)


def create_tabs(names_to_panels: Dict[str, Panel]):
    tabs = QTabWidget()
    for panel_name, panel in names_to_panels.items():
        tabs.addTab(panel, panel_name)
    return tabs


class APIUsageCostDialog(QDialog):
    def __init__(self, html_content, parent=None):
        super(APIUsageCostDialog, self).__init__(parent, Qt.WindowType.WindowCloseButtonHint)
        self.setWindowTitle("API Usage Cost")

        # Layout to organize widgets
        layout = QVBoxLayout()

        # QLabel to display HTML content
        label = QTextEdit()
        label.setReadOnly(True)
        label.setHtml(f'<style>{CSS}</style>{html_content}')

        layout.addWidget(label)

        # QPushButton to close the dialog
        self.close_button = QPushButton("Close")
        self.close_button.clicked.connect(self.close)
        layout.addWidget(self.close_button)

        self.setLayout(layout)
        self.setStyleSheet(HTMLPOPUP_STYLE)
        self.resize(400, 300)


class PysideApp(QMainWindow, BaseApp):
    send_text_signal = Signal(str, PanelNames)
    send_panel_continue_signal = Signal(PanelNames)
    a_application = None

    def __init__(self, mutex, condition, step_runner=None):
        super().__init__()
        self.products: Dict[Stage, Any] = {}
        self.popups = set()
        self.api_usage_cost = StageToCost()

        self.panels = {
            PanelNames.SYSTEM_PROMPT: EditableTextPanel("System Prompt", "", ("Default",)),
            PanelNames.MISSION_PROMPT: EditableTextPanel("Mission Prompt", "", ("Default",)),
            PanelNames.RESPONSE: EditableTextPanel("Response", "", ()),
            PanelNames.PRODUCT: EditableTextPanel("Product", "", ()),
            PanelNames.FEEDBACK: EditableTextPanel("Feedback", "", ("AI Review", "No comments")),
        }
        central_widget = QWidget()
        self.layout = QHBoxLayout(central_widget)

        self.setStyleSheet(APP_STYLE)

        # Left side is a VBox with "Continue" button above and the steps panel below
        left_side = QVBoxLayout()
        self.layout.addLayout(left_side)

        # Steps panel
        self.step_panel = StepsPanel()
        left_side.addWidget(self.step_panel)

        # Right side is a QVBoxLayout with a header on top and a splitter with the text panels below
        right_side = QVBoxLayout()
        self.layout.addLayout(right_side)

        header_and_checkbox = QHBoxLayout()
        right_side.addLayout(header_and_checkbox)

        # Header:
        self.header = QLabel()
        self.header.setTextFormat(Qt.TextFormat.RichText)
        self.header.setSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Fixed)
        self.header.setStyleSheet(f"color: {CURRENT_STEP_COLOR}; font-size: 24px; font-weight: bold;")

        spacer = QSpacerItem(0, 0, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        header_and_checkbox.addWidget(self.header)
        header_and_checkbox.addItem(spacer)

        # Checkboxes:
        check_boxes = QVBoxLayout()
        header_and_checkbox.addLayout(check_boxes)

        # Bypass-continue checkbox:
        self.bypass_continue_checkbox = QCheckBox("Auto continue")
        self.bypass_continue_checkbox.setChecked(False)
        self.bypass_continue_checkbox.setToolTip(
            "When unchecked, a user 'Continue' approval is required for each LLM iteration.\n"
            "When checked, the app only stops where explicit user choices are required.")
        self.bypass_continue_checkbox.setStyleSheet(QCHECKBOX_STYLE)
        check_boxes.addWidget(self.bypass_continue_checkbox)

        # Bypass-mission prompt checkbox:
        self.bypass_mission_prompt_checkbox = QCheckBox("Auto mission prompt")
        self.bypass_mission_prompt_checkbox.setChecked(False)
        self.bypass_mission_prompt_checkbox.setToolTip(
            "When unchecked, user can edit each mission prompt.\n"
            "When checked, the default mission prompt is automatically used.")
        self.bypass_mission_prompt_checkbox.setStyleSheet(QCHECKBOX_STYLE)
        check_boxes.addWidget(self.bypass_mission_prompt_checkbox)

        # add button with $ sign that opens the pricing dialog, displaying the api usage cost per stage
        self.api_usage_cost_button = QPushButton("")
        self.api_usage_cost_button.setFixedWidth(130)
        self._update_api_usage_cost_button()
        self.api_usage_cost_button.clicked.connect(self.show_api_usage_cost_dialog)
        header_and_checkbox.addWidget(self.api_usage_cost_button)

        # Splitter with the text panels
        main_splitter = QSplitter(Qt.Orientation.Horizontal)
        left_splitter = QSplitter(Qt.Orientation.Vertical)
        right_splitter = QSplitter(Qt.Orientation.Vertical)
        main_splitter.addWidget(left_splitter)
        main_splitter.addWidget(right_splitter)

        # Add the panels to the splitters (the top-right panel is a tab widget)
        self.tabs = create_tabs({'Response': self.panels[PanelNames.RESPONSE],
                                 'Product': self.panels[PanelNames.PRODUCT]})
        self.tabs.setStyleSheet(TABS_STYLE)
        left_splitter.addWidget(self.panels[PanelNames.SYSTEM_PROMPT])
        left_splitter.addWidget(self.panels[PanelNames.MISSION_PROMPT])
        right_splitter.addWidget(self.tabs)
        right_splitter.addWidget(self.panels[PanelNames.FEEDBACK])
        left_splitter.setSizes([100, 500])
        right_side.setStretchFactor(main_splitter, 1)

        main_splitter.setStyleSheet(MAIN_SPLITTER_STYLE)
        right_side.addWidget(main_splitter)

        self.layout.addLayout(right_side)
        self.setCentralWidget(central_widget)

        self.resize(1000, 600)
        self.setWindowTitle("data-to-paper")

        # Worker thread setup
        self.worker = Worker(mutex, condition)
        # Slot now accepts a string argument for the initial text
        self.worker.request_panel_continue_signal.connect(self.upon_request_panel_continue)
        self.worker.request_text_signal.connect(self.upon_request_text)
        self.worker.show_text_signal.connect(self.upon_show_text)
        self.worker.set_focus_on_panel_signal.connect(self.upon_set_focus_on_panel)
        self.worker.advance_stage_int_signal.connect(self.upon_advance_stage_int)
        self.worker.send_product_of_stage_signal.connect(self.upon_send_product_of_stage)
        self.worker.set_status_signal.connect(self.upon_set_status)
        self.worker.set_header_signal.connect(self.upon_set_header)
        self.worker.send_api_usage_cost_signal.connect(self.upon_send_api_usage_cost)

        # Define the request_text and show_text methods
        self.request_panel_continue = self.worker.worker_request_panel_continue
        self.request_text = self.worker.worker_request_text
        self.show_text = self.worker.worker_show_text
        self.set_focus_on_panel = self.worker.worker_set_focus_on_panel
        self.send_product_of_stage = self.worker.worker_send_product_of_stage
        self._set_status = self.worker.worker_set_status
        self.set_header = self.worker.worker_set_header
        self.send_api_usage_cost = self.worker.worker_send_api_usage_cost

        # Connect UI elements
        for panel_name in PanelNames:
            if panel_name == PanelNames.PRODUCT:
                continue
            self.panels[panel_name].submit_button.clicked.connect(partial(self.submit_text, panel_name=panel_name))
            self.panels[panel_name].continue_button.clicked.connect(partial(self.upon_panel_continue,
                                                                            panel_name=panel_name))

        # Connect the MainWindow signal to the worker's slot
        self.send_text_signal.connect(self.worker.receive_text_signal)
        self.send_panel_continue_signal.connect(self.worker.receive_panel_continue_signal)

    def closeEvent(self, event):
        """Override the closeEvent to gracefully stop the worker thread."""
        print("Main window is closing...")
        for popup in self.popups:
            popup.close()
        self.stage_to_reset_to = True
        # imitate a button click if resetting when waiting for user input
        self._initiate_panel_button_clicks()
        if not self.worker.wait(5000):
            print("Worker thread did not finish in time. Forcefully terminating...")
        event.accept()

    @classmethod
    def get_instance(cls):
        if cls.instance is None:
            mutex = QMutex()
            condition = QWaitCondition()
            cls.instance = cls(mutex, condition)
        return cls.instance

    def advance_stage(self, stage: Union[Stage, int, bool]):
        if isinstance(stage, Stage):
            stage = list(self._get_stages()).index(stage)
        elif stage is True:  # completed
            stage = len(self._get_stages())
        elif stage is False:  # error
            stage = -1
        self.worker.worker_advance_stage_int(stage)

    def start_worker(self, func_to_run=None):
        # Start the worker thread
        self.worker.func_to_run = func_to_run or self._run_all_steps
        self.worker.start()

    def initialize(self, func_to_run=None):
        self.step_panel.init_ui(
            {(stage.value, stage.name, stage.resettable): (
                partial(self.show_product_for_stage, stage),
                partial(self.confirm_and_perform_reset_to_stage, stage)
            ) for stage in self._get_stages()})
        self.show()
        self.start_worker(func_to_run)
        return get_or_create_q_application_if_app_is_pyside().exec()

    @Slot(PanelNames, int, str)
    def upon_set_status(self, panel_name: PanelNames, position: int, status: str = ''):
        if panel_name == PanelNames.PRODUCT or panel_name == PanelNames.RESPONSE:
            panel_name = [PanelNames.PRODUCT, PanelNames.RESPONSE]
        else:
            panel_name = [panel_name]
        for name in panel_name:
            if position == 1:
                self.panels[name].set_header_right(status)
            else:
                self.panels[name].set_header(status)
            self.panels[name].update()

    @Slot(str)
    def upon_set_header(self, header: str):
        self.header.setText(header)

    def show_product_for_stage(self, stage: Stage):
        """
        Open a popup window to show the product of a stage.
        """
        product_text = self.products.get(stage, '<span style="color: white;">Not created yet.</span>')
        if product_text.startswith('<a href="file://'):
            # open the file in the normal OS application
            file_path = product_text.split('"')[1]
            open_file_on_os(file_path)
            return
        popup = HtmlPopup(stage.value, product_text)
        popup.show()
        self.popups.add(popup)
        popup.finished.connect(self.popup_closed)

    def _update_api_usage_cost_button(self):
        cost = self.api_usage_cost.get_total_cost()
        self.api_usage_cost_button.setText(f"API Cost: ${cost:.2f}")

    def show_api_usage_cost_dialog(self):
        """
        Open a popup window to show the pricing of the API usage per stage.
        """
        popup = APIUsageCostDialog(self.api_usage_cost.as_html())
        popup.show()
        self.popups.add(popup)
        popup.finished.connect(self.popup_closed)

    def popup_closed(self):
        closed_popup = self.sender()
        self.popups.discard(closed_popup)

    @Slot(PanelNames)
    def upon_request_panel_continue(self, panel_name: PanelNames):
        if self.bypass_continue_checkbox.isChecked():
            self.send_panel_continue_signal.emit(panel_name)
        else:
            panel = self.panels[panel_name]
            panel.wait_for_continue()

    @Slot(PanelNames, str, str, dict)
    def upon_request_text(self, panel_name: PanelNames, initial_text: str = '',
                          title: Optional[str] = None,
                          instructions: Optional[str] = None,
                          in_field_instructions: Optional[str] = None,
                          optional_suggestions: Dict[str, str] = None):
        panel = self.panels[panel_name]
        if optional_suggestions is None:
            optional_suggestions = {}
        if panel_name == PanelNames.MISSION_PROMPT and self.bypass_mission_prompt_checkbox.isChecked():
            self.send_text_signal.emit(panel_name, initial_text)
            return
        panel.edit_text(initial_text, title, instructions, in_field_instructions, list(optional_suggestions.values()))

    @Slot(PanelNames)
    def upon_panel_continue(self, panel_name: PanelNames):
        self.send_panel_continue_signal.emit(panel_name)

    @Slot(PanelNames)
    def submit_text(self, panel_name: PanelNames):
        panel = self.panels[panel_name]
        text = panel.get_text()
        self.send_text_signal.emit(panel_name, text)

    @Slot(PanelNames, str, bool, bool)
    def upon_show_text(self, panel_name: PanelNames, text: str, is_html: bool = False,
                       scroll_to_bottom: bool = False):
        panel = self.panels[panel_name]
        panel.set_text(text, is_html)
        if scroll_to_bottom:
            panel.scroll_to_bottom()

    @Slot(PanelNames)
    def upon_set_focus_on_panel(self, panel_name: PanelNames):
        panel = self.panels[panel_name]
        panel.text_edit.setFocus()
        # if the panel is in a tab, switch to the tab
        for i in range(self.tabs.count()):
            if self.tabs.widget(i) == panel:
                self.tabs.setCurrentIndex(i)
                break

    @Slot(Stage)
    def upon_advance_stage_int(self, stage: int):
        self.step_panel.set_step_by_index(stage)

    @Slot(Stage, str)
    def upon_send_product_of_stage(self, stage: Stage, product_text: str):
        self.products[stage] = product_text

    @Slot(object)
    def upon_send_api_usage_cost(self, stages_to_costs: StageToCost):
        self.api_usage_cost = stages_to_costs
        self._update_api_usage_cost_button()

    def confirm_and_perform_reset_to_stage(self, stage: Stage):
        dialog = QDialog(self)
        dialog.setWindowTitle("Reset to Step")
        layout = QVBoxLayout(dialog)
        label = QLabel(f"Are you sure you want to reset to the '{stage.value}' step?\n"
                       f"This will irrevocably delete all progress after this step.")
        layout.addWidget(label)
        buttons_layout = QHBoxLayout()
        layout.addLayout(buttons_layout)
        yes_button = QPushButton("Yes")
        yes_button.clicked.connect(partial(self.perform_reset_to_stage, stage, dialog))
        no_button = QPushButton("No")
        no_button.clicked.connect(dialog.close)
        buttons_layout.addWidget(yes_button)
        buttons_layout.addWidget(no_button)
        dialog.exec()

    def perform_reset_to_stage(self, stage: Stage, dialog: Optional[QDialog] = None):
        if dialog is not None:
            dialog.close()

        self.stage_to_reset_to = stage

        # delete all product for stages after the reset stage
        stage_index = stage.get_index()
        stages = list(self._get_stages())
        for stage in stages[stage_index:]:
            self.products.pop(stage, None)
        # imitate a button click if resetting when waiting for user input
        self._initiate_panel_button_clicks()

    def _initiate_panel_button_clicks(self):
        for panel in self.panels.values():
            for button in [panel.submit_button, panel.continue_button]:
                if button.isVisible():
                    button.click()



from dataclasses import dataclass, field
from typing import Optional, Dict

from data_to_paper.code_and_output_files.code_and_output import CodeAndOutput
from data_to_paper.base_products import Products, NameDescriptionStageGenerator
from data_to_paper.base_products.file_descriptions import DataFileDescriptions

from .stage import DemoStages


@dataclass
class DemoProducts(Products):
    """
    All the products generated in the demo.
    """
    data_file_descriptions: DataFileDescriptions = field(default_factory=DataFileDescriptions)
    research_goal: Optional[str] = None
    code_and_output: CodeAndOutput = None
    paper_sections: Dict[str, str] = field(default_factory=dict)

    def _get_generators(self) -> Dict[str, NameDescriptionStageGenerator]:
        return {
            **super()._get_generators(),

            'data_file_descriptions': NameDescriptionStageGenerator(
                'Data File Descriptions',
                '{}',
                DemoStages.DATA,
                lambda: self.data_file_descriptions,
            ),

            'research_goal': NameDescriptionStageGenerator(
                'Research Goal',
                '{}',
                DemoStages.GOAL,
                lambda: self.research_goal,
            ),

            'code_and_output': NameDescriptionStageGenerator(
                'Code and Output',
                '{description}',
                DemoStages.CODE,
                lambda: {
                    'description': self.code_and_output.to_text(with_header=False)},
            ),

            'title_and_abstract': NameDescriptionStageGenerator(
                'Title and Abstract',
                "```latex\n{}\n\n{}```",
                DemoStages.WRITING,
                lambda: (self.paper_sections['title'],
                         self.paper_sections['abstract']),
            ),
        }



from pathlib import Path


def convert_to_latex_comment(text: str) -> str:
    """
    Convert a string to a comment in latex.
    """
    lines = text.split('\n')
    return '% ' + '\n% '.join(lines)


def convert_filename_to_label(filename, label) -> str:
    """
    Convert a filename to a label.
    """
    if label:
        raise ValueError(f'Do not provide the `label` argument. The label is derived from the filename.')
    if not filename:
        return ''
    label = Path(filename).stem
    ext = Path(filename).suffix
    if ext:
        raise ValueError(f'Invalid filename: "{filename}". The filename must not have an extension.')

    # check if the label is valid:
    if not label.isidentifier():
        raise ValueError(f'Invalid filename: "{filename}". The filename must be a valid identifier.')
    label = label.replace('_', '-')
    return label



import colorama

from data_to_paper.terminate.exceptions import TerminateException


def resource_checking(title: str):
    def _resource_checking(func):

        def wrapper(*args, **kwargs):
            from data_to_paper.text.highlighted_text import colored_text, format_text_with_code_blocks
            from data_to_paper.utils.print_to_file import IS_LOGGING_ENABLED
            print(f"\n*** {title}\nRunning...")
            try:
                with IS_LOGGING_ENABLED.temporary_set(False):
                    func(*args, **kwargs)
            except Exception as e:
                color = colorama.Fore.CYAN if isinstance(e, TerminateException) else colorama.Fore.RED
                print(colored_text(f"Test failed:", color=colorama.Fore.RED))
                print(format_text_with_code_blocks(str(e), text_color=color))
            else:
                print(colored_text("Test successful!", color=colorama.Fore.GREEN))

        return wrapper
    return _resource_checking



import re
import difflib

import numpy as np


def word_count(text: str) -> int:
    """
    Count the number of words in provided test.
    """
    return len(re.findall(r'\w+', text))


def line_count(text: str) -> int:
    """
    Count the number of lines in provided test.
    """
    return len(text.splitlines())


def is_bulleted_list(text: str) -> bool:
    """
    Check if the provided text is a bulleted list, i.e. that there are lines starting with '- ' or '* '.
    """
    return any(line.startswith('- ') or line.startswith('* ') for line in text.splitlines())


def diff_strs(str1: str, str2: str, context: int = 1,
              add_template: str = '\033[91m{}\033[0m ',
              remove_template: str = '\033[92m{}\033[0m ',
              ) -> str:
    """
    Return a string that shows the diff between two strings.
    `context` is the number of words to show before and after a diff.
    `add_template` and `remove_template` are the templates to use for added and removed words.
    """
    d = difflib.Differ()
    diff = list(d.compare(str1.split(), str2.split()))
    is_diff = np.array([word[0] != ' ' for word in diff])

    # to_show is a True/False flag that indicates whether to show the diff or not. it is True if there is a diff, or
    # we are within a distance `context` from a diff.
    to_show = is_diff.copy()
    for i in range(context):
        to_show[1:] |= to_show[:-1]
        to_show[:-1] |= to_show[1:]

    s = ''
    three_dots = False
    for word, show in zip(diff, to_show):
        if show:
            tag, word = word[0], word[2:]
            if tag == ' ':  # no change
                s += word + ' '
            elif tag == '-':  # removed
                s += remove_template.format(word)
            elif tag == '+':  # added
                s += add_template.format(word)
            elif tag == '?':  # ?
                s += f'[{word}] '
            else:
                raise ValueError(f'Unknown tag: {tag}')
            three_dots = False
        else:
            if not three_dots:
                s += ' ... '
                three_dots = True
    return s



import pandas as pd

from data_to_paper.research_types.hypothesis_testing.coding.analysis.my_utils.df_to_latex import analysis_df_to_latex


def displayitems_df_to_latex(df: pd.DataFrame, filename: str, **kwargs):
    """
    Same as analysis_df_to_latex, but for displayitems.
    See docstring of analysis_df_to_latex for more information.
    """
    return analysis_df_to_latex(df, filename, **kwargs)



from dataclasses import dataclass
from typing import Optional

from data_to_paper.exceptions import data_to_paperException
from data_to_paper.text import wrap_as_block


@dataclass
class FailedToExtractLatexContent(data_to_paperException, ValueError):
    """
    Raised when the latex content could not be extracted from the response.
    """
    reason: str

    def __str__(self):
        return self.reason


@dataclass
class BaseLatexProblemInCompilation(data_to_paperException, ValueError):
    pass


@dataclass
class LatexProblemInCompilation(BaseLatexProblemInCompilation):
    """
    Raised when the latex content could not be compiled.
    """
    latex_content: str
    pdflatex_output: str
    problem_starting_term: str

    def _extract_error_message(self) -> str:
        """
        Extract the error message from the pdflatex output.
        """
        lines = self.pdflatex_output.splitlines()
        first_line_of_error_message = next((i for i, line in enumerate(lines) if
                                            line.startswith(self.problem_starting_term)), None)
        return '\n'.join(lines[first_line_of_error_message:first_line_of_error_message + 4])

    @property
    def error_message(self) -> str:
        """
        Get the error message from the pdflatex output.
        """
        return wrap_as_block(self._extract_error_message(), 'error')

    def get_latex_exception_line_number(self) -> Optional[int]:
        """
        Get the line number of the latex exception.
        """
        error_message = self._extract_error_message()
        if '\nl.' not in error_message:
            return None
        return int(error_message.split('\nl.')[1].split(' ')[0]) - 1  # -1 because the latex line numbers start at 1

    def _get_erroneous_lines(self) -> Optional[str]:
        """
        Get the erroneous lines from the latex content.
        """
        error_line = self.get_latex_exception_line_number()
        if error_line is None:
            return None
        return '\n'.join(self.latex_content.splitlines()[error_line - 1:error_line + 2])

    def __str__(self):
        erroneous_lines = self._get_erroneous_lines()
        if erroneous_lines is None:
            message_header = f'There was a latex compilation problem.\n\n'
        else:
            message_header = f'There was a latex compilation problem in these lines:\n{erroneous_lines}\n\n'
        return f'{message_header}Got the following pdflatex error:\n{self.error_message}'


@dataclass
class LatexCompilationError(LatexProblemInCompilation):
    """
    Raised when the latex content could not be compiled.
    """
    problem_starting_term: str = '! '


@dataclass
class LatexNumCommandFormulaEvalError(BaseLatexProblemInCompilation):
    """
    Raised when there is an error evaluating the latex \num command.
    """
    expression: str
    exception: Exception

    def __str__(self):
        return f'Failed to evaluate the latex \\num command for the expression:\n{self.expression}\n' \
               f'Got the following exception:\n{self.exception}'


@dataclass
class LatexNumCommandNoExplanation(BaseLatexProblemInCompilation):
    """
    Raised when latex \num command missing an explanation.
    """
    expression: str

    def __str__(self):
        return (f'The latex \\num command must have two arguments: a formula and an explanation. '
                f'Expected format: \\num{{<formula>, "explanation"}}, but got:\n{self.expression}')


@dataclass
class LatexNestedNumCommandError(BaseLatexProblemInCompilation):
    """
    Raised when there is an error evaluating the latex \num command.
    """
    expression: str

    def __str__(self):
        return f'Nested \\num commands are not supported. In:\n{self.expression}'


@dataclass
class PlainNumberLatexNumCommandError(BaseLatexProblemInCompilation):
    """
    Raised when the latex \num command contains a plain number.
    """
    expression: str

    def __str__(self):
        return (f'The latex \\num command should only be used with a formula. '
                f'It should not be used for a plain number. In:\n{self.expression}')


@dataclass
class UnwantedCommandsUsedInLatex(data_to_paperException, ValueError):
    """
    Raised when the latex content contains unwanted commands.
    """
    unwanted_commands: list

    def __str__(self):
        return f'Unwanted commands used in latex:\n{self.unwanted_commands}\n'


@dataclass
class NonLatexCitations(data_to_paperException, ValueError):
    """
    Raised when there are citations that are not written using latex \\cite{} command.
    """
    non_latex_citations: list

    def __str__(self):
        return f'The following citations are not written using latex \\cite{{}} command:\n' \
               f'{self.non_latex_citations}\n\n' \
               f'Please use latex \\cite{{}} command to write citations.\n'



from dataclasses import dataclass
from typing import Any


@dataclass
class SerializableValue:
    """
    Base class for a serializable value.
    """

    value: Any

    def _serialize_value(self):
        return self.value

    @classmethod
    def _start_with(cls) -> str:
        return f'{cls.__name__}: '

    def serialize(self):
        return f'{self._start_with()}{self._serialize_value()}'

    @classmethod
    def from_serialized_value(cls, serialized_value: str):
        return cls(serialized_value)

    @classmethod
    def deserialize(cls, serialized: str):
        if serialized.startswith(cls._start_with()):
            return cls.from_serialized_value(serialized[len(cls._start_with()):])
        raise ValueError(f'Unknown serialized object: {serialized}')


def get_all_subclasses(cls):
    all_subclasses = []

    for subclass in cls.__subclasses__():
        all_subclasses.append(subclass)
        all_subclasses.extend(get_all_subclasses(subclass))

    return all_subclasses


def deserialize_serializable_value(serialized: str) -> SerializableValue:
    """
    Deserializes a serialized object.
    """
    # iterate over all subclasses AND their subclasses
    for cls in get_all_subclasses(SerializableValue):
        try:
            return cls.deserialize(serialized)
        except ValueError:
            pass
    raise ValueError(f'Unknown serialized object: {serialized}')



from dataclasses import dataclass
from typing import Tuple

from data_to_paper.base_steps import LatexReviewBackgroundProductsConverser
from data_to_paper.text import dedent_triple_quote_str


from .cast import DemoAgent
from .products import DemoProducts


@dataclass
class WriteTitleAndAbstract(LatexReviewBackgroundProductsConverser):
    """
    Base class for the writer of a paper section in latex format.
    """
    conversation_name: str = 'Title and Abstract'
    products: DemoProducts = None
    background_product_fields: Tuple[str, ...] = ('data_file_descriptions', 'research_goal', 'code_and_output')
    product_fields_from_which_response_is_extracted: Tuple[str, ...] = None
    allow_citations_from_step: str = None
    should_remove_citations_from_section: bool = False

    max_reviewing_rounds: int = 1
    goal_noun: str = 'Title and Abstract for a funny math article'
    goal_verb: str = 'write'
    performer: str = 'scientific writer'
    reviewer: str = 'scientific reviewer'
    assistant_agent: DemoAgent = DemoAgent.Performer
    user_agent: DemoAgent = DemoAgent.Writer
    journal_name: str = 'Crazy Math Journal'
    termination_phrase: str = 'This is just great and funny.'

    system_prompt: str = dedent_triple_quote_str("""
        You are a writer.

        You will write a fake funny scientific article for the journal {journal_name}.
        """)

    mission_prompt: str = dedent_triple_quote_str("""
        Based on the material provided above ({actual_background_product_names}), \t
        please {goal_verb} only the {goal_noun} for a {journal_name} article.
        Do not write any other parts!

        While making it funny, please make sure to specifically relate to the specific numerical results that we have.

        Your response should be formatted as {your_response_should_be_formatted_as}
        """)

    your_response_should_be_formatted_as: str = dedent_triple_quote_str("""
        a triple-backtick block of latex text, like this:
        ```latex
        \\title{<Your title here>}
        \\begin{abstract}
        <Your abstract here>
        \\end{abstract}
        ```
        """)

    other_system_prompt: str = dedent_triple_quote_str("""
        You are a reviewer for a writer who is writing a funny scientific-like paper about their results.
        Your job is to provide constructive bullet-point feedback. 
        If you feel that the writing does not need further improvements, you should reply only with:
        "{termination_phrase}".
        """)

    sentence_to_add_at_the_end_of_reviewer_response: str = dedent_triple_quote_str("""\n\n
        Please correct your response according to any points in my feedback that you find relevant and applicable.
        Send back a complete rewrite of the {pretty_section_names}.
        Make sure to send the full corrected {pretty_section_names}, not just the parts that were revised.
        """)

    sentence_to_add_at_the_end_of_performer_response: str = dedent_triple_quote_str("""
        Please provide a bullet-point list of constructive feedback on the above {pretty_section_names} \t
        for my paper. Do not provide positive feedback, only provide actionable instructions for improvements in \t
        bullet points. 
        In particular, make sure that the section is correctly grounded in the information provided above, \t
        yet is written in a funny way.
        If you find any inconsistencies or discrepancies, please mention them explicitly in your feedback.

        You should only provide feedback on the {pretty_section_names}. Do not provide feedback on other sections \t
        or other parts of the paper, like code output or Python code, provided above.

        If you don't see any flaws, respond solely with "{termination_phrase}".

        IMPORTANT: You should EITHER provide bullet-point feedback, or respond solely with "{termination_phrase}"; \t
        If you chose to provide bullet-point feedback then DO NOT include "{termination_phrase}".
        """)
